# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from typing import Callable, Dict, Optional, Tuple, Union

import jax.numpy as jnp
import numpy as np
from jax.typing import ArrayLike

from pyhgf.model import Network
from pyhgf.response import first_level_binary_surprise, first_level_gaussian_surprise


class HGF(Network):
    """The two-level and three-level Hierarchical Gaussian Filters (HGF).

    This class uses pre-made node structures that correspond to the most widely used
    HGF. The inputs can be continuous or binary.

    Attributes
    ----------
    model_type :
        The model implemented (can be `"continuous"`, `"binary"` or `"custom"`).
    n_levels :
        The number of hierarchies in the model, including the input vector. It cannot be
        less than 2.
    .. note::
        The parameter structure also incorporates the value and volatility coupling
        strength with children and parents (i.e. `"value_coupling_parents"`,
        `"value_coupling_children"`, `"volatility_coupling_parents"`,
        `"volatility_coupling_children"`).

    """

    def __init__(
        self,
        n_levels: Optional[int] = 2,
        model_type: str = "continuous",
        initial_mean: Dict = {
            "1": 0.0,
            "2": 0.0,
            "3": 0.0,
        },
        initial_precision: Dict = {
            "1": 1.0,
            "2": 1.0,
            "3": 1.0,
        },
        continuous_precision: Union[float, np.ndarray, ArrayLike] = 1e4,
        tonic_volatility: Dict = {
            "1": -3.0,
            "2": -3.0,
            "3": -3.0,
        },
        volatility_coupling: Dict = {"1": 1.0, "2": 1.0},
        eta0: Union[float, np.ndarray, ArrayLike] = 0.0,
        eta1: Union[float, np.ndarray, ArrayLike] = 1.0,
        binary_precision: Union[float, np.ndarray, ArrayLike] = jnp.inf,
        tonic_drift: Dict = {
            "1": 0.0,
            "2": 0.0,
            "3": 0.0,
        },
    ) -> None:
        r"""Parameterization of the HGF model.

        Parameters
        ----------
        n_levels :
            The number of hierarchies in the perceptual model (can be `2` or `3`). If
            `None`, the nodes hierarchy is not created and might be provided afterwards.
            Defaults to `2` for a 2-level HGF.
        model_type : str
            The model type to use (can be `"continuous"` or `"binary"`).
        initial_mean :
            A dictionary containing the initial values for the initial mean at
            different levels of the hierarchy. Defaults set to `0.0`.
        initial_precision :
            A dictionary containing the initial values for the initial precision at
            different levels of the hierarchy. Defaults set to `1.0`.
        continuous_precision :
            The expected precision of the continuous input node. Default to `1e4`. Only
            relevant if `model_type="continuous"`.
        tonic_volatility :
            A dictionary containing the initial values for the tonic volatility
            at different levels of the hierarchy. This represents the tonic
            part of the variance (the part that is not affected by the parent node).
            Defaults are set to `-3.0`.
        volatility_coupling :
            A dictionary containing the initial values for the volatility coupling
            at different levels of the hierarchy. This represents the phasic part of
            the variance (the part that is affected by the parent nodes) and will
            define the strength of the connection between the node and the parent
            node. Defaults set to `1.0`.
        eta0 :
            The first categorical value of the binary node. Defaults to `0.0`. Only
            relevant if `model_type="binary"`.
        eta1 :
            The second categorical value of the binary node. Defaults to `0.0`. Only
            relevant if `model_type="binary"`.
        binary_precision :
            The precision of the binary input node. Default to `jnp.inf`. Only relevant
            if `model_type="binary"`.
        tonic_drift :
            A dictionary containing the initial values for the tonic drift
            at different levels of the hierarchy. This represents the drift of the
            random walk. Defaults set all entries to `0.0` (no drift).

        """
        Network.__init__(self)
        self.model_type = model_type
        self.n_levels = n_levels

        if model_type not in ["continuous", "binary"]:
            raise ValueError("Invalid model type.")
        else:
            if model_type == "continuous":
                # X - 0
                self.add_nodes(
                    precision=continuous_precision,
                )

                # X - 1
                self.add_nodes(
                    value_children=([0], [1.0]),
                    node_parameters={
                        "mean": initial_mean["1"],
                        "precision": initial_precision["1"],
                        "tonic_volatility": tonic_volatility["1"],
                        "tonic_drift": tonic_drift["1"],
                    },
                )

                # X - 2
                self.add_nodes(
                    volatility_children=([1], [volatility_coupling["1"]]),
                    node_parameters={
                        "mean": initial_mean["2"],
                        "precision": initial_precision["2"],
                        "tonic_volatility": tonic_volatility["2"],
                        "tonic_drift": tonic_drift["2"],
                    },
                )

                #########################
                # Meta volatility level #
                #########################
                if self.n_levels == 3:
                    self.add_nodes(
                        volatility_children=([2], [volatility_coupling["2"]]),
                        node_parameters={
                            "mean": initial_mean["3"],
                            "precision": initial_precision["3"],
                            "tonic_volatility": tonic_volatility["3"],
                            "tonic_drift": tonic_drift["3"],
                        },
                    )

            elif model_type == "binary":

                if binary_precision == jnp.inf:
                    # X - 0
                    self.add_nodes(
                        kind="binary-state",
                        node_parameters={
                            "mean": initial_mean["1"],
                            "precision": initial_precision["1"],
                        },
                    )

                # X - 1
                self.add_nodes(
                    kind="continuous-state",
                    value_children=([0], [1.0]),
                    node_parameters={
                        "mean": initial_mean["2"],
                        "precision": initial_precision["2"],
                        "tonic_volatility": tonic_volatility["2"],
                        "tonic_drift": tonic_drift["2"],
                    },
                )

                #########################
                # Meta volatility level #
                #########################
                if self.n_levels == 3:
                    self.add_nodes(
                        volatility_children=([1], [volatility_coupling["2"]]),
                        node_parameters={
                            "mean": initial_mean["3"],
                            "precision": initial_precision["3"],
                            "tonic_volatility": tonic_volatility["3"],
                            "tonic_drift": tonic_drift["3"],
                        },
                    )

            # initialize the model so it is ready to receive new observations
            self.create_belief_propagation_fn()

    def surprise(
        self,
        response_function: Optional[Callable] = None,
        response_function_inputs: Tuple = (),
        response_function_parameters: Optional[
            Union[np.ndarray, ArrayLike, float]
        ] = None,
    ) -> float:
        """Surprise of the model conditioned by the response function.

        The surprise (negative log probability) depends on the input data, the model
        parameters, the response function, its inputs and its additional parameters
        (optional).

        Parameters
        ----------
        response_function :
            The response function to use to compute the model surprise. If `None`
            (default), return the sum of Gaussian surprise if `model_type=="continuous"`
            or the sum of the binary surprise if `model_type=="binary"`.
        response_function_inputs :
            A list of tuples with the same length as the number of models. Each tuple
            contains additional data and parameters that can be accessible to the
            response functions.
        response_function_parameters :
            A list of additional parameters that will be passed to the response
            function. This can include values over which inferece is performed in a
            PyMC model (e.g. the inverse temperature of a binary softmax).

        Returns
        -------
        surprise :
            The model's surprise given the input data and the response function.

        """
        if response_function is None:
            response_function = (
                first_level_gaussian_surprise
                if self.model_type == "continuous"
                else first_level_binary_surprise
            )

        return response_function(
            hgf=self,
            response_function_inputs=response_function_inputs,
            response_function_parameters=response_function_parameters,
        )



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from typing import Dict

import jax.numpy as jnp

from pyhgf.math import Normal
from pyhgf.typing import Attributes, Edges


def dirichlet_node_prediction(
    edges: Edges,
    attributes: Dict,
    node_idx: int,
    **args,
) -> Attributes:
    """Prediction of a Dirichlet process node.

    Parameters
    ----------
    edges :
        The edges of the neural network as a tuple of
        :py:class:`pyhgf.typing.Indexes`. The tuple has the same length as node number.
        For each node, the index lists the value/volatility parents/children.
    attributes :
        The attributes of the probabilistic nodes.
    node_idx :
        Pointer to the Dirichlet process input node.

    Returns
    -------
    attributes :
        The attributes of the probabilistic nodes.
    edges :
        The edges of the neural network.
    input_nodes_idx :
        Static input nodes' parameters for the neural network.
    dirichlet_node :
        Static parameters of the Dirichlet process node.

    """
    # get the parameter (mean and variance) from the EF-normal parent nodes
    value_parent_idxs = edges[node_idx].value_parents
    if value_parent_idxs is not None:
        parameters = jnp.array(
            [
                Normal().parameters(xis=attributes[parent_idx]["xis"])
                for parent_idx in value_parent_idxs
            ]
        )

        attributes[node_idx]["expected_means"] = parameters[:, 0]
        attributes[node_idx]["expected_sigmas"] = jnp.sqrt(parameters[:, 1])

    return attributes



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

import pkgutil
from io import BytesIO
from typing import Tuple, Union

import numpy as np
import pandas as pd

__version__ = "0.2.0"


def load_data(dataset: str) -> Union[Tuple[np.ndarray, ...], np.ndarray]:
    """Load dataset for continuous or binary HGF.

    Parameters
    ----------
    dataset : str
        The type of data to load. Can be `"continous"` or `"binary"`.

    Returns
    -------
    data : np.ndarray
        The data (a 1d timeseries).

    Notes
    -----
    The continuous time series is the standard USD-CHF conversion rates over time used
    in the Matlab examples.

    The binary dataset is from Iglesias et al. (2013) [#] (see the full dataset
    `here <https://www.research-collection.ethz.ch/handle/20.500.11850/454711)>`_. The
    binary set consist of one vector *u*, the observations, and one vector *y*, the
    decisions.

    References
    ----------
    .. [#] Iglesias, S., Kasper, L., Harrison, S. J., Manka, R., Mathys, C., & Stephan,
      K. E. (2021). Cholinergic and dopaminergic effects on prediction error and
      uncertainty responses during sensory associative learning. In NeuroImage (Vol.
      226, p. 117590). Elsevier BV. https://doi.org/10.1016/j.neuroimage.2020.117590

    """
    if dataset == "continuous":
        data = pd.read_csv(
            BytesIO(pkgutil.get_data(__name__, "data/usdchf.txt")),  # type: ignore
            names=["x"],
        ).x.to_numpy()
    elif dataset == "binary":
        u = pd.read_csv(
            BytesIO(
                pkgutil.get_data(__name__, "data/binary_input.txt")  # type: ignore
            ),
            names=["x"],
        ).x.to_numpy()
        y = pd.read_csv(
            BytesIO(
                pkgutil.get_data(__name__, "data/binary_response.txt")  # type: ignore
            ),
            names=["x"],
        ).x.to_numpy()
        data = (u, y)
    else:
        raise ValueError("Invalid dataset argument. Should be 'continous' or 'binary'.")

    return data



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

import jax.numpy as jnp
import numpy as np

from pyhgf import load_data
from pyhgf.model import HGF, Network


def test_plotting_functions():
    # Read USD-CHF data
    timeserie = load_data("continuous")

    ##############
    # Continuous #
    # ------------

    # Set up standard 2-level HGF for continuous inputs
    two_level_continuous = HGF(
        n_levels=2,
        model_type="continuous",
        initial_mean={"1": 1.04, "2": 1.0},
        initial_precision={"1": 1e4, "2": 1e1},
        tonic_volatility={"1": -13.0, "2": -2.0},
        tonic_drift={"1": 0.0, "2": 0.0},
        volatility_coupling={"1": 1.0},
    ).input_data(input_data=timeserie)

    # plot trajectories
    two_level_continuous.plot_trajectories()

    # plot correlations
    two_level_continuous.plot_correlations()

    # plot node structures
    two_level_continuous.plot_network()

    # plot nodes
    two_level_continuous.plot_nodes(
        node_idxs=2,
        show_posterior=True,
    )

    # Set up standard 3-level HGF for continuous inputs
    three_level_continuous = HGF(
        n_levels=3,
        model_type="continuous",
        initial_mean={"1": 1.04, "2": 1.0, "3": 1.0},
        initial_precision={"1": 1e4, "2": 1e1, "3": 1e1},
        tonic_volatility={"1": -13.0, "2": -2.0, "3": -2.0},
        tonic_drift={"1": 0.0, "2": 0.0, "3": 0.0},
        volatility_coupling={"1": 1.0, "2": 1.0},
    ).input_data(input_data=timeserie)

    # plot trajectories
    three_level_continuous.plot_trajectories()

    # plot correlations
    three_level_continuous.plot_correlations()

    # plot node structures
    three_level_continuous.plot_network()

    # plot nodes
    three_level_continuous.plot_nodes(
        node_idxs=2,
        show_posterior=True,
    )

    ##########
    # Binary #
    # --------

    # Read binary input
    u, _ = load_data("binary")

    two_level_binary_hgf = HGF(
        n_levels=2,
        model_type="binary",
        initial_mean={"1": 0.0, "2": 0.5},
        initial_precision={"1": 0.0, "2": 1e4},
        tonic_volatility={"1": None, "2": -6.0},
        tonic_drift={"1": None, "2": 0.0},
        volatility_coupling={"1": None},
        eta0=0.0,
        eta1=1.0,
        binary_precision=jnp.inf,
    ).input_data(u)

    # plot trajectories
    two_level_binary_hgf.plot_trajectories()

    # plot correlations
    two_level_binary_hgf.plot_correlations()

    # plot node structures
    two_level_binary_hgf.plot_network()

    # plot node structures
    two_level_binary_hgf.plot_nodes(
        node_idxs=1,
        show_posterior=True,
    )

    three_level_binary_hgf = HGF(
        n_levels=3,
        model_type="binary",
        initial_mean={"1": 0.0, "2": 0.5, "3": 0.0},
        initial_precision={"1": 0.0, "2": 1e4, "3": 1e1},
        tonic_volatility={"1": None, "2": -6.0, "3": -2.0},
        tonic_drift={"1": None, "2": 0.0, "3": 0.0},
        volatility_coupling={"1": None, "2": 1.0},
        eta0=0.0,
        eta1=1.0,
        binary_precision=jnp.inf,
    ).input_data(u)

    # plot trajectories
    three_level_binary_hgf.plot_trajectories()

    # plot correlations
    three_level_binary_hgf.plot_correlations()

    # plot node structures
    three_level_binary_hgf.plot_network()

    # plot node structures
    three_level_binary_hgf.plot_nodes(
        node_idxs=2,
        show_posterior=True,
    )

    #############
    # Categorical
    # -----------

    # generate some categorical inputs data
    input_data = np.array(
        [np.random.multinomial(n=1, pvals=[0.1, 0.2, 0.7]) for _ in range(3)]
    ).T
    input_data = np.vstack([[0.0] * input_data.shape[1], input_data])

    # create the categorical HGF
    categorical_hgf = Network().add_nodes(
        kind="categorical-state",
        node_parameters={
            "n_categories": 3,
            "binary_parameters": {"tonic_volatility_2": -2.0},
        },
    )

    # fitting the model forwards
    categorical_hgf.input_data(
        input_data=(input_data, np.ones(input_data.shape, dtype=int))
    )

    # plot node structures
    categorical_hgf.plot_network()



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

import numpy as np

from pyhgf import load_data
from pyhgf.model import HGF
from pyhgf.response import binary_softmax, binary_softmax_inverse_temperature


def test_binary_responses():
    u, y = load_data("binary")

    # two-level binary HGF
    # --------------------
    two_level_binary_hgf = HGF(
        n_levels=2,
        model_type="binary",
        initial_mean={"1": 0.5, "2": 0.0},
        initial_precision={"1": 0.0, "2": 1.0},
        tonic_volatility={"2": -6.0},
    ).input_data(input_data=u)

    # binary sofmax
    # -------------
    surprise = two_level_binary_hgf.surprise(
        response_function=binary_softmax, response_function_inputs=y
    )
    assert np.isclose(surprise.sum(), 195.81573)

    # binary sofmax with inverse temperature
    # --------------------------------------
    surprise = two_level_binary_hgf.surprise(
        response_function=binary_softmax_inverse_temperature,
        response_function_inputs=y,
        response_function_parameters=2.0,
    )
    assert np.isclose(surprise.sum(), 188.77818)



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from typing import TYPE_CHECKING

import jax.numpy as jnp
from jax.typing import ArrayLike

from pyhgf.math import binary_surprise, gaussian_surprise

if TYPE_CHECKING:
    from pyhgf.model import HGF


def first_level_gaussian_surprise(
    hgf: "HGF", response_function_inputs=None, response_function_parameters=None
) -> float:
    """Gaussian surprise at the first level of a probabilistic network.

    .. note::
      The Gaussian surprise at the first level is the default method to compute surprise
      for continuous models. The function returns `jnp.inf` if the model could not fit
      at a given time point.

    Parameters
    ----------
    hgf :
        An instance of the HGF model.
    response_function_inputs :
        The inputs to the response functions, not required here.
    response_function_parameters :
        Additionnal parameters for the response function, not required here.

    Returns
    -------
    surprise :
        The model's surprise given the input data.

    """
    # compute the sum of Gaussian surprise at the first level
    # the input value at time t is compared to the Gaussian prediction at t-1
    surprise = gaussian_surprise(
        x=hgf.node_trajectories[0]["mean"],
        expected_mean=hgf.node_trajectories[1]["expected_mean"],
        expected_precision=hgf.node_trajectories[1]["expected_precision"],
    )

    # Return an infinite surprise if the model could not fit at any point
    return jnp.where(jnp.isnan(surprise), jnp.inf, surprise)


def total_gaussian_surprise(
    hgf: "HGF", response_function_inputs=None, response_function_parameters=None
) -> float:
    """Sum of the Gaussian surprise across the probabilistic network.

    .. note::
      The function returns `jnp.inf` if the model could not fit at a given time point.

    Parameters
    ----------
    hgf :
        An instance of the HGF model.
    response_function_inputs :
        The inputs to the response functions, not required here.
    response_function_parameters :
        Additionnal parameters for the response function, not required here.

    Returns
    -------
    surprise :
        The model's surprise given the input data.

    """
    # compute the sum of Gaussian surprise across every node
    surprise = 0.0

    # then we do the same for every node that is not an input node
    # and not the parent of an input node
    for i in range(len(hgf.edges)):
        if hgf.edges[i].node_type == 2:
            surprise += gaussian_surprise(
                x=hgf.node_trajectories[i]["mean"],
                expected_mean=hgf.node_trajectories[i]["expected_mean"],
                expected_precision=hgf.node_trajectories[i]["expected_precision"],
            )

    # Return an infinite surprise if the model could not fit at any point
    return jnp.where(
        jnp.any(jnp.isnan(hgf.node_trajectories[1]["mean"])), jnp.inf, surprise
    )


def first_level_binary_surprise(
    hgf: "HGF", response_function_inputs=None, response_function_parameters=None
) -> float:
    """Time series of binary surprises for all binary state nodes.

    .. note::
      The binary surprise is the default method to compute surprise when
      `model_type=="binary"`, therefore this method will only return the valid time
      points, and `jnp.inf` if the model could not fit.

    Parameters
    ----------
    hgf :
        An instance of the HGF model.
    response_function_inputs :
        The inputs to the response functions, not required here.
    response_function_parameters :
        Additionnal parameters for the response function, not required here.

    Returns
    -------
    surprise :
        The model's surprise given the input data.

    """
    surprise = jnp.zeros(len(hgf.node_trajectories[-1]["time_step"]))
    for binary_input_idx in hgf.input_idxs:
        surprise += binary_surprise(
            expected_mean=hgf.node_trajectories[binary_input_idx]["expected_mean"],
            x=hgf.node_trajectories[binary_input_idx]["mean"],
        )

    # Return an infinite surprise if the model cannot fit
    surprise = jnp.where(
        jnp.isnan(surprise),
        jnp.inf,
        surprise,
    )

    return surprise


def binary_softmax(
    hgf: "HGF",
    response_function_inputs=ArrayLike,
    response_function_parameters=ArrayLike,
) -> float:
    """Surprise under the binary sofmax model.

    Parameters
    ----------
    hgf :
        An instance of the HGF model.
    response_function_inputs :
        The inputs to the response functions, here containing the decision from the
        paraticipant at time *k* [0 or 1].
    response_function_parameters :
        Additionnal parameters for the response function, not required here.

    Returns
    -------
    surprise :
        The surprise under the binary sofmax model.

    """
    # the expected values at the first level of the HGF
    beliefs = hgf.node_trajectories[0]["expected_mean"]

    # the binary surprises
    surprise = binary_surprise(x=response_function_inputs, expected_mean=beliefs)

    # ensure that inf is returned if the model cannot fit
    surprise = jnp.where(jnp.isnan(surprise), jnp.inf, surprise)

    return surprise


def binary_softmax_inverse_temperature(
    hgf: "HGF",
    response_function_inputs=ArrayLike,
    response_function_parameters=ArrayLike,
) -> float:
    r"""Surprise from a binary sofmax parametrized by the inverse temperature.

    The probability of chosing A is given by:

    .. math::

       P(A|\hat{\mu}^{(k)_{1}, t) = \frac{1}{1+e^{-t\hat{\mu}^{(k)_{1}}}

    Where :math:`\hat{mu}^{(k)_{1}` is the expected probability of A at the firt level,
    and :math:`t` is the temperature parameter.

    Parameters
    ----------
    hgf :
        An instance of the HGF model.
    response_function_inputs :
        The inputs to the response functions, here containing the decision from the
        paraticipant at time *k* [0 or 1].
    response_function_parameters :
        Additionnal parameters for the response function (optional). Here, the inverse
        temperature  is provided.

    Returns
    -------
    surprise :
        The surprise under the binary sofmax model.

    """
    # the expected values at the first level of the HGF
    beliefs = (
        hgf.node_trajectories[0]["expected_mean"] ** response_function_parameters
    ) / (
        hgf.node_trajectories[0]["expected_mean"] ** response_function_parameters
        + (1 - hgf.node_trajectories[0]["expected_mean"])
        ** response_function_parameters
    )

    # the binary surprises
    surprise = binary_surprise(x=response_function_inputs, expected_mean=beliefs)

    # ensure that inf is returned if the model cannot fit
    surprise = jnp.where(jnp.isnan(surprise), jnp.inf, surprise)

    return surprise



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

import jax.numpy as jnp

from pyhgf import load_data
from pyhgf.math import binary_surprise, gaussian_density, sigmoid
from pyhgf.model import Network
from pyhgf.utils import beliefs_propagation


def test_gaussian_density():
    surprise = gaussian_density(
        x=jnp.array([1.0, 1.0]),
        mean=jnp.array([0.0, 0.0]),
        precision=jnp.array([1.0, 1.0]),
    )
    assert jnp.all(jnp.isclose(surprise, 0.24197073))


def test_sgm():
    assert jnp.all(jnp.isclose(sigmoid(jnp.array([0.3, 0.3])), 0.5744425))


def test_binary_surprise():
    surprise = binary_surprise(
        x=jnp.array([1.0]),
        expected_mean=jnp.array([0.2]),
    )
    assert jnp.all(jnp.isclose(surprise, 1.609438))


def test_update_binary_input_parents():

    binary_hgf = (
        Network()
        .add_nodes(kind="binary-state")
        .add_nodes(value_children=0, mean=1.0, tonic_volatility=1.0)
        .add_nodes(volatility_children=1, mean=1.0, tonic_volatility=1.0)
        .create_belief_propagation_fn()
    )

    data = jnp.ones(1)
    time_steps = jnp.ones(1)
    observed = jnp.ones(1)

    # apply sequence
    new_attributes, _ = beliefs_propagation(
        attributes=binary_hgf.attributes,
        inputs=(data, time_steps, observed),
        update_sequence=binary_hgf.update_sequence,
        edges=binary_hgf.edges,
        input_idxs=(0,),
    )
    for idx, val in zip(
        ["mean", "expected_mean", "expected_precision"],
        [1.0, 0.7310586, 0.19661193],
    ):
        assert jnp.isclose(new_attributes[0][idx], val)
    for idx, val in zip(
        ["mean", "expected_mean", "precision", "expected_precision"],
        [1.8515793, 1.0, 0.31581485, 0.11920292],
    ):
        assert jnp.isclose(new_attributes[1][idx], val)
    for idx, val in zip(
        ["mean", "expected_mean", "precision", "expected_precision"],
        [0.12210602, 1.0, 0.47702926, 0.26894143],
    ):
        assert jnp.isclose(new_attributes[2][idx], val)


def test_binary_scan_loop():

    u, _ = load_data("binary")

    binary_hgf = (
        Network()
        .add_nodes(kind="binary-state")
        .add_nodes(value_children=0, mean=0.0, tonic_volatility=1.0)
        .add_nodes(volatility_children=1, mean=0.0, tonic_volatility=1.0)
        .input_data(input_data=u)
    )

    for idx, val in zip(
        ["mean", "expected_mean", "precision", "expected_precision"],
        [-1.5600492, -1.5068374, 3.4091694, 3.2606704],
    ):
        assert jnp.isclose(binary_hgf.node_trajectories[1][idx][-1], val)
    for idx, val in zip(
        ["mean", "expected_mean", "precision", "expected_precision"],
        [-5.35091, -5.320357, 0.025971143, 0.024351958],
    ):
        assert jnp.isclose(binary_hgf.node_trajectories[2][idx][-1], val)



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from functools import partial
from typing import Dict

from jax import jit

from pyhgf.math import sigmoid
from pyhgf.typing import Edges


@partial(jit, static_argnames=("edges", "node_idx"))
def binary_state_node_prediction(
    attributes: Dict, edges: Edges, node_idx: int, **args
) -> Dict:
    r"""Get the new expected mean and precision of a binary state node.

    The predictions of a binary state node :math:`b` at time :math:`k` depends on the
    prediction of its value parent :math:`a`, such as:

    .. math::

        \hat{\mu}_b^{(k)} = \frac{1}{1 + e^{-\hat{\mu}_a^{(k)}}}

    and

    .. math::

        \hat{\pi}_b^{(k)} = \frac{1}{\hat{\mu}^{(k)}(1-\hat{\mu}^{(k)})}

    Parameters
    ----------
    attributes :
        The attributes of the probabilistic nodes.
    edges :
        The edges of the probabilistic nodes as a tuple of
        :py:class:`pyhgf.typing.Indexes`. The tuple has the same length as the node
        number. For each node, the index lists the value and volatility parents and
        children.
    time_step :
        The interval between the previous time point and the current time point.
    node_idx :
        Pointer to the binary state node.

    Returns
    -------
    attributes :
        The attributes of the probabilistic nodes.

    """
    # List the (unique) value parent of the binary state node
    expected_mean = 0.0
    for value_parent_idx in edges[node_idx].value_parents:  # type: ignore
        expected_mean += attributes[value_parent_idx]["expected_mean"]

    # Estimate the new expected mean using the sigmoid transform
    expected_mean = sigmoid(expected_mean)

    # Estimate the new expected precision from the new expected mean - use the inverse
    # to fit the posterior update of the value parent(eq. 97, Weber et al., v1)
    # here we also update the precision
    # so the expected precision remains during the observation step
    attributes[node_idx]["expected_precision"] = expected_mean * (1 - expected_mean)
    attributes[node_idx]["precision"] = attributes[node_idx]["expected_precision"]
    attributes[node_idx]["expected_mean"] = expected_mean

    return attributes



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

import jax.numpy as jnp
from pytest import raises

from pyhgf import load_data
from pyhgf.model import Network
from pyhgf.typing import AdjacencyLists
from pyhgf.utils import list_branches


def test_imports():
    """Test the data import function"""
    _ = load_data("continuous")
    _, _ = load_data("binary")

    with raises(Exception):
        load_data("error")


def test_add_edges():
    """Test the add_edges function."""
    network = Network().add_nodes().add_nodes(n_nodes=3)
    with raises(Exception):
        network.add_edges(kind="error")

    network.add_edges(
        kind="volatility", parent_idxs=2, children_idxs=0, coupling_strengths=1
    )
    network.add_edges(parent_idxs=1, children_idxs=0, coupling_strengths=1.0)


def test_find_branch():
    """Test the find_branch function."""
    edges = (
        AdjacencyLists(0, (1,), None, None, None, (None,)),
        AdjacencyLists(2, None, (2,), (0,), None, (None,)),
        AdjacencyLists(2, None, None, None, (1,), (None,)),
        AdjacencyLists(2, (4,), None, None, None, (None,)),
        AdjacencyLists(2, None, None, (3,), None, (None,)),
    )
    branch_list = list_branches([0], edges, branch_list=[])
    assert branch_list == [0, 1, 2]


def test_set_update_sequence():
    """Test the set_update_sequence function."""

    # a standard binary HGF
    network1 = (
        Network()
        .add_nodes(kind="binary-state")
        .add_nodes(value_children=0)
        .create_belief_propagation_fn()
    )

    predictions, updates = network1.update_sequence
    assert len(predictions) == 2
    assert len(updates) == 2

    # a standard continuous HGF
    network2 = (
        Network()
        .add_nodes()
        .add_nodes(value_children=0)
        .add_nodes(volatility_children=1)
        .create_belief_propagation_fn(update_type="standard")
    )
    predictions, updates = network2.update_sequence
    assert len(predictions) == 3
    assert len(updates) == 4

    # a generic input with a normal-EF node
    network3 = (
        Network()
        .add_nodes(kind="generic-state")
        .add_nodes(kind="ef-normal", value_children=0)
        .create_belief_propagation_fn()
    )
    predictions, updates = network3.update_sequence
    assert len(predictions) == 0
    assert len(updates) == 2

    # a Dirichlet node
    network4 = (
        Network()
        .add_nodes(kind="generic-state")
        .add_nodes(kind="DP-state", value_children=0, alpha=0.1, batch_size=2)
        .add_nodes(
            kind="ef-normal",
            n_nodes=2,
            value_children=1,
            xis=jnp.array([0.0, 1 / 8]),
            nus=15.0,
        )
        .create_belief_propagation_fn()
    )
    predictions, updates = network4.update_sequence
    assert len(predictions) == 1
    assert len(updates) == 4



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from functools import partial
from typing import Dict

from jax import Array, jit

from pyhgf.typing import Edges


@partial(jit, static_argnames=("node_idx"))
def continuous_node_value_prediction_error(
    attributes: Dict,
    node_idx: int,
) -> Array:
    r"""Compute the value prediction error of a state node.

    The value prediction error :math:`\delta_j^{(k)}` of a continuous state node is
    given by:

    .. math::

        \delta_j^{(k)} = \mu_j^{k} - \hat{\mu}_j^{k}

    Parameters
    ----------
    attributes :
        The attributes of the probabilistic nodes.
    node_idx :
        Pointer to the value parent node that will be updated.

    Returns
    -------
    attributes :
        The attributes of the probabilistic nodes.

    See Also
    --------
    continuous_node_volatility_prediction_error, continuous_node_prediction_error

    References
    ----------
    .. [1] Weber, L. A., Waade, P. T., Legrand, N., Møller, A. H., Stephan, K. E., &
       Mathys, C. (2023). The generalized Hierarchical Gaussian Filter (Version 1).
       arXiv. https://doi.org/10.48550/ARXIV.2305.10937

    """
    # value prediction error
    value_prediction_error = (
        attributes[node_idx]["mean"] - attributes[node_idx]["expected_mean"]
    )

    # divide by the number of value parents
    if attributes[node_idx]["value_coupling_parents"] is not None:
        value_prediction_error /= len(attributes[node_idx]["value_coupling_parents"])

    # send to the value parent node for later use in the update step
    attributes[node_idx]["temp"]["value_prediction_error"] = value_prediction_error

    return attributes


@partial(jit, static_argnames=("node_idx"))
def continuous_node_volatility_prediction_error(
    attributes: Dict, node_idx: int
) -> Dict:
    r"""Compute the volatility prediction error of a state node.

    The volatility prediction error :math:`\Delta_j^{(k)}` of a state node
    :math:`j` is given by:

    .. math::

        \Delta_j^{(k)} = \frac{\hat{\pi}_j^{(k)}}{\pi_j^{(k)}} +
        \hat{\pi}_j^{(k)} \left( \delta_j^{(k)} \right)^2 - 1

    Parameters
    ----------
    attributes :
        The attributes of the probabilistic nodes.
    node_idx :
        Pointer to the node that will be updated.

    Returns
    -------
    attributes :
        The attributes of the probabilistic nodes.

    See Also
    --------
    continuous_node_value_prediction_error, continuous_node_prediction_error

    References
    ----------
    .. [1] Weber, L. A., Waade, P. T., Legrand, N., Møller, A. H., Stephan, K. E., &
       Mathys, C. (2023). The generalized Hierarchical Gaussian Filter (Version 1).
       arXiv. https://doi.org/10.48550/ARXIV.2305.10937

    """
    # compute the volatility prediction error (VOPE)
    volatility_prediction_error = (
        (attributes[node_idx]["expected_precision"] / attributes[node_idx]["precision"])
        + attributes[node_idx]["expected_precision"]
        * (attributes[node_idx]["temp"]["value_prediction_error"]) ** 2
        - 1
    )

    # divide by the number of volatility parents
    if attributes[node_idx]["volatility_coupling_parents"] is not None:
        volatility_prediction_error /= len(
            attributes[node_idx]["volatility_coupling_parents"]
        )

    attributes[node_idx]["temp"][
        "volatility_prediction_error"
    ] = volatility_prediction_error

    return attributes


@partial(jit, static_argnames=("edges", "node_idx"))
def continuous_node_prediction_error(
    attributes: Dict, node_idx: int, edges: Edges, **args
) -> Dict:
    """Store prediction errors in an input node.

    Parameters
    ----------
    attributes :
        The attributes of the probabilistic nodes.
    .. note::
        The parameter structure also incorporates the value and volatility coupling
        strength with children and parents (i.e. `"value_coupling_parents"`,
        `"value_coupling_children"`, `"volatility_coupling_parents"`,
        `"volatility_coupling_children"`).
    node_idx :
        Pointer to the continuous node.
    edges :
        The edges of the probabilistic nodes as a tuple of
        :py:class:`pyhgf.typing.Indexes`. The tuple has the same length as the node
        number. For each node, the index lists the value and volatility parents and
        children.

    Returns
    -------
    attributes :
        The updated attributes of the probabilistic nodes.

    See Also
    --------
    continuous_node_volatility_prediction_error, continuous_node_value_prediction_error

    References
    ----------
    .. [1] Weber, L. A., Waade, P. T., Legrand, N., Møller, A. H., Stephan, K. E., &
       Mathys, C. (2023). The generalized Hierarchical Gaussian Filter (Version 1).
       arXiv. https://doi.org/10.48550/ARXIV.2305.10937

    """
    # Store value prediction errors
    # -----------------------------
    attributes = continuous_node_value_prediction_error(
        attributes=attributes, node_idx=node_idx
    )

    # Store volatility prediction errors
    # ----------------------------------
    attributes = continuous_node_volatility_prediction_error(
        attributes=attributes, node_idx=node_idx
    )

    return attributes



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from functools import partial
from typing import Dict

from jax import jit


@partial(jit, static_argnames=("node_idx"))
def set_observation(
    attributes: Dict,
    node_idx: int,
    values: float,
    observed: int,
) -> Dict:
    r"""Add observations to the target node by setting the posterior to a given value.

    Parameters
    ----------
    attributes :
        The attributes of the probabilistic network.
    node_idx :
        Pointer to the input node.
    values :
        The new observed value.
    observed :
        Whether value was observed or not.

    Returns
    -------
    attributes :
        The attributes of the probabilistic network.

    """
    attributes[node_idx]["mean"] = values
    attributes[node_idx]["observed"] = observed

    return attributes



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from functools import partial
from typing import Dict

import jax.numpy as jnp
from jax import jit

from pyhgf.math import binary_surprise, dirichlet_kullback_leibler
from pyhgf.typing import Edges


@partial(jit, static_argnames=("edges", "node_idx"))
def categorical_state_update(
    attributes: Dict, node_idx: int, edges: Edges, **args
) -> Dict:
    """Update the categorical input node given an array of binary observations.

    This function should be called **after** the update of the implied binary HGFs. It
    receives a `None` as the boolean observations are passed to the binary inputs
    directly. This update uses the expected probability of the binary input to update
    an implied Dirichlet distribution.

    Parameters
    ----------
    attributes :
        The attributes of the probabilistic nodes.
    .. note::
        `"psis"` is the value coupling strength. It should have the same length as the
        volatility parents' indexes. `"volatility_coupling"` is the volatility coupling
        strength. It should have the same length as the volatility parents' indexes.
    node_idx :
        Pointer to the node that needs to be updated.
    edges :
        The edges of the probabilistic nodes as a tuple of
        :py:class:`pyhgf.typing.Indexes`. The tuple has the same length as the node
        number. For each node, the index lists the value and volatility parents and
        children.

    Returns
    -------
    attributes :
        The updated parameters structure.

    See Also
    --------
    binary_input_update, continuous_input_update

    """
    # get the expected values before the update
    expected_mean = jnp.array(
        [
            attributes[value_parent_idx]["expected_mean"]
            for value_parent_idx in edges[node_idx].value_parents  # type: ignore
        ]
    )

    # get the new values after the update from the continuous state
    updated_mean = jnp.array(
        [
            attributes[edges[parent_idx].value_parents[0]]["mean"]
            for parent_idx in edges[node_idx].value_parents  # type: ignore
        ]
    )
    updated_mean = 1 / (1 + jnp.exp(-updated_mean))  # logit transform

    # compute the prediction error (observed - expected) at time K
    pe = attributes[node_idx]["mean"] - expected_mean

    # the differential of expectations (parent posterior - parents expectation)
    delta_xi = updated_mean - expected_mean

    # using the new PE, we can update nu and the alpha vector
    nu = (pe / delta_xi) - 1
    alpha = (nu * expected_mean) + 1  # concentration parameters for the Dirichlet

    # in case alpha contains NaNs
    # alpha = jnp.where(jnp.isnan(alpha), 1.0, alpha)

    # compute Bayesian surprise as :
    # 1 - KL divergence from the concentration parameters
    attributes[node_idx]["kl_divergence"] = dirichlet_kullback_leibler(
        attributes[node_idx]["alpha"], alpha
    )
    # 2 - the sum of binary surprises observed in the parents nodes
    attributes[node_idx]["surprise"] = jnp.sum(
        binary_surprise(x=attributes[node_idx]["mean"], expected_mean=expected_mean)
    )

    # save the new concentration parameters
    attributes[node_idx]["alpha"] = alpha
    attributes[node_idx]["mean"] = updated_mean

    # prediction mean
    # attributes[node_idx]["mean"] = alpha / jnp.sum(alpha)

    return attributes



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from functools import partial
from typing import Dict

from jax import jit

from pyhgf.typing import Edges


@partial(jit, static_argnames=("edges", "node_idx"))
def categorical_state_prediction_error(
    attributes: Dict, node_idx: int, edges: Edges, **args
) -> Dict:
    """Prediction error from a categorical state node.

    The update will pass the input observations to the binary state nodes.

    Parameters
    ----------
    attributes :
        The attributes of the probabilistic nodes.
    .. note::
        `"psis"` is the value coupling strength. It should have the same length as the
        volatility parents' indexes. `"volatility_coupling"` is the volatility coupling
        strength. It should have the same length as the volatility parents' indexes.
    node_idx :
        Pointer to the node that needs to be updated.
    edges :
        The edges of the probabilistic nodes as a tuple of
        :py:class:`pyhgf.typing.Indexes`. The tuple has the same length as the node
        number. For each node, the index lists the value and volatility parents and
        children.

    Returns
    -------
    attributes :
        The updated parameters structure.

    See Also
    --------
    binary_input_update, continuous_input_update

    """
    # pass the mean to the binary state nodes
    for mean, observed, value_parent_idx in zip(
        attributes[node_idx]["mean"],
        attributes[node_idx]["observed"],
        edges[node_idx].value_parents,  # type: ignore
    ):
        attributes[value_parent_idx]["mean"] = mean
        attributes[value_parent_idx]["observed"] = observed

    return attributes



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from typing import Callable, Dict, NamedTuple, Optional, Tuple, Union


class AdjacencyLists(NamedTuple):
    """Indexes to a node's value and volatility parents.

    The variable `node_type` encode the type of state node:
    * 0: input node.
    * 1: binary state node.
    * 2: continuous state node.
    * 3: exponential family state node - univariate Gaussian distribution with unknown
        mean and unknown variance.
    * 4: Dirichlet Process state node.

    The variable `coupling_fn` list the coupling functions between this nodes and the
    children nodes. If `None` is provided, a linear coupling is assumed.

    """

    node_type: int
    value_parents: Optional[Tuple]
    volatility_parents: Optional[Tuple]
    value_children: Optional[Tuple]
    volatility_children: Optional[Tuple]
    coupling_fn: Tuple[Optional[Callable], ...]


# the nodes' attributes
Attributes = Dict[Union[int, str], Dict]

# the network edges
Edges = Tuple[AdjacencyLists, ...]

# the update sequence
Sequence = Tuple[Tuple[int, Callable], ...]
UpdateSequence = Tuple[Sequence, Sequence]

# a fully defined network
NetworkParameters = Tuple[Attributes, Edges, UpdateSequence]



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>
import jax.numpy as jnp

from pyhgf.math import (
    MultivariateNormal,
    Normal,
    binary_surprise_finite_precision,
    gaussian_predictive_distribution,
    gaussian_surprise,
    sigmoid_inverse_temperature,
)


def test_gaussian_surprise():
    surprise = gaussian_surprise(
        x=jnp.array([1.0, 1.0]),
        expected_mean=jnp.array([0.0, 0.0]),
        expected_precision=jnp.array([1.0, 1.0]),
    )
    assert jnp.all(jnp.isclose(surprise, 1.4189385))


def test_multivariate_normal():

    ss = MultivariateNormal.sufficient_statistics(jnp.array([1.0, 2.0]))
    assert jnp.isclose(ss, jnp.array([1.0, 2.0, 1.0, 2.0, 4.0], dtype="float32")).all()

    bm = MultivariateNormal.base_measure(2)
    assert bm == 0.15915494309189535


def test_normal():

    ss = Normal.sufficient_statistics(jnp.array(1.0))
    assert jnp.isclose(ss, jnp.array([1.0, 1.0], dtype="float32")).all()

    bm = Normal.base_measure()
    assert bm == 0.3989423

    ess = Normal.expected_sufficient_statistics(mu=0.0, sigma=1.0)
    assert jnp.isclose(ess, jnp.array([0.0, 1.0], dtype="float32")).all()

    par = Normal.parameters(xis=[5.0, 29.0])
    assert jnp.isclose(jnp.array(par), jnp.array([5.0, 4.0], dtype="float32")).all()


def test_gaussian_predictive_distribution():

    pdf = gaussian_predictive_distribution(x=1.5, xi=[0.0, 1 / 8], nu=5.0)
    assert jnp.isclose(pdf, jnp.array(0.00845728, dtype="float32"))


def test_binary_surprise_finite_precision():

    surprise = binary_surprise_finite_precision(
        value=1.0,
        expected_mean=0.0,
        expected_precision=1.0,
        eta0=0.0,
        eta1=1.0,
    )
    assert surprise == 1.4189385


def test_sigmoid_inverse_temperature():
    s = sigmoid_inverse_temperature(x=0.4, temperature=6.0)
    assert jnp.isclose(s, jnp.array(0.08070617906683485, dtype="float32"))



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from functools import partial
from typing import Callable, List, Optional, Union

import jax.numpy as jnp
import numpy as np
import pytensor.tensor as pt
from jax import Array, grad, jit, vmap
from jax.tree_util import Partial
from jax.typing import ArrayLike
from pytensor.graph import Apply, Op
from pytensor.tensor.variable import TensorVariable

from pyhgf.model import HGF


@partial(jit, static_argnames=("hgf", "response_function"))
def logp(
    mean_1: float,
    mean_2: float,
    mean_3: float,
    precision_1: float,
    precision_2: float,
    precision_3: float,
    tonic_volatility_1: float,
    tonic_volatility_2: float,
    tonic_volatility_3: float,
    tonic_drift_1: float,
    tonic_drift_2: float,
    tonic_drift_3: float,
    volatility_coupling_1: float,
    volatility_coupling_2: float,
    input_precision: float,
    response_function_parameters: ArrayLike,
    input_data: ArrayLike,
    time_steps: ArrayLike,
    response_function_inputs: Optional[ArrayLike],
    response_function: Optional[Callable],
    hgf: HGF,
) -> float:
    """Compute the log-probability of a decision model under belief trajectories.

    This function returns the evidence of a single Hierarchical Gaussian Filter given
    network parameters, input data and behaviours under a decision model.

    Parameters
    ----------
    mean_1 :
        The mean at the first level of the HGF. For the continuous HGF, this is the mean
        of the first value parent (x_1). For the binary HGF this is the mean of the
        binary state node (x_0).
    mean_2 :
        The mean at the second level of the HGF. For the continuous HGF, this is the
        mean of the first volatility parent (x_2). For the binary HGF this is the mean
        of the first continuous state node (x_1).
    mean_3 :
        The mean at the third level of the HGF. The value of this parameter will be
        ignored when using a two-level HGF (`n_levels=2`). For the continuous HGF, this
        is the mean of the second volatility parent (x_3). For the binary HGF this is
        the mean of the first volatility parent (x_2).
    precision_1 :
        The precision at the first level of the HGF. For the continuous HGF, this is the
        precision of the first value parent (x_1). For the binary HGF this is the
        precision of the binary state node (x_0).
    precision_2 :
        The precision at the second level of the HGF. For the continuous HGF, this is
        the precision of the first volatility parent (x_2). For the binary HGF this is
        the precision of the first continuous state node (x_1).
    precision_3 :
        The precision at the third level of the HGF. The value of this parameter will
        be ignored when using a two-level HGF (`n_levels=2`). For the continuous HGF,
        this is the precision of the second volatility parent (x_3). For the binary HGF
        this is the precision of the first volatility parent (x_2).
    tonic_volatility_1 :
        The tonic volatility at the first level (x_1 for the continuous HGF, x_2 for the
        binary HGF). This parameter represents the tonic part of the variance (the part
        that is not inherited from parent nodes).
    tonic_volatility_2 :
        The tonic volatility at the second level (x_2 for the continuous HGF, x_3 for
        the binary HGF). This parameter represents the tonic part of the variance (the
        part that is not inherited from parent nodes).
    tonic_volatility_3 :
        The tonic volatility at the third level of the HGF. This parameter represents
        the tonic part of the variance (the part that is not inherited from parent
        nodes). This parameter is only used for a three-level continuous HGF.
    tonic_drift_1 :
        The tonic drift at the first level of the HGF (x_1 for the continuous HGF,
        x_2 for the binary HGF). This parameter represents the drift of the random walk.
    tonic_drift_2 :
        The tonic drift at the second level of the HGF (x_2 for the continuous HGF,
        x_3 for the binary HGF). This parameter represents the drift of the random walk.
    tonic_drift_3 :
        The tonic drift at the third level of the HGF. This parameter represents the
        drift of the random walk. This parameter is only used for a three-level
        continuous HGF.
    volatility_coupling_1 :
        The volatility coupling between the first and second levels of the HGF (between
        x_1 and x_2 for a continuous HGF, and between x_2 and x_3 for a binary HGF).
        This represents the phasic part of the variance (the part affected by the
        parent nodes). Defaults to `1.0` (full connectivity).
    volatility_coupling_2 :
        The volatility coupling between the second and third levels of the HGF (x_2 and
        x_2 for a continuous HGF, not applicable to a binary HGF). This represents the
        phasic part of the variance (the part affected by the parent nodes). Defaults
        to `1.0`  (full connectivity). The value of this parameter will be ignored when
        using a two-level HGF (`n_levels=2`).
    input_precision :
        The expected precision associated with the continuous input.
    response_function_parameters :
        An array of additional parameters that will be passed to the response function
        to compute the surprise. This can include values over which inference is
        performed in a PyMC model (e.g. the inverse temperature of a binary softmax).
    input_data :
        An array of input time series where the first dimension is the number of models
        to fit in parallel.
    time_steps :
        An array of input time steps where the first dimension is the number of models
        to fit in parallel.
    response_function_inputs :
        An array of behavioural inputs passed to the response function where the first
        dimension is the number of models to fit in parallel.
    response_function :
        The response function that is used by the decision model.
    hgf :
        An instance of a two or three-level Hierarchical Gaussian Filter.

    Returns
    -------
    logp :
        The log-probability (negative surprise).

    """
    if hgf.model_type == "continuous":

        # update this network's attributes
        hgf.attributes[0]["precision"] = input_precision

        hgf.attributes[1]["mean"] = mean_1
        hgf.attributes[2]["mean"] = mean_2

        hgf.attributes[1]["precision"] = precision_1
        hgf.attributes[2]["precision"] = precision_2

        hgf.attributes[1]["tonic_volatility"] = tonic_volatility_1
        hgf.attributes[2]["tonic_volatility"] = tonic_volatility_2

        hgf.attributes[1]["tonic_drift"] = tonic_drift_1
        hgf.attributes[2]["tonic_drift"] = tonic_drift_2

        hgf.attributes[1]["volatility_coupling_parents"] = (volatility_coupling_1,)
        hgf.attributes[2]["volatility_coupling_children"] = (volatility_coupling_1,)

        if hgf.n_levels == 3:
            hgf.attributes[3]["mean"] = mean_3
            hgf.attributes[3]["precision"] = precision_3
            hgf.attributes[3]["tonic_volatility"] = tonic_volatility_3
            hgf.attributes[3]["tonic_drift"] = tonic_drift_3
            hgf.attributes[2]["volatility_coupling_parents"] = (volatility_coupling_2,)
            hgf.attributes[3]["volatility_coupling_children"] = (volatility_coupling_2,)

    elif hgf.model_type == "binary":

        # update this network's attributes
        hgf.attributes[0]["mean"] = mean_1
        hgf.attributes[1]["mean"] = mean_2

        hgf.attributes[0]["precision"] = precision_1
        hgf.attributes[1]["precision"] = precision_2

        hgf.attributes[1]["tonic_volatility"] = tonic_volatility_2

        hgf.attributes[1]["tonic_drift"] = tonic_drift_1

        if hgf.n_levels == 3:
            hgf.attributes[2]["mean"] = mean_3
            hgf.attributes[2]["precision"] = precision_3
            hgf.attributes[2]["tonic_volatility"] = tonic_volatility_3
            hgf.attributes[2]["tonic_drift"] = tonic_drift_3
            hgf.attributes[1]["volatility_coupling_parents"] = (volatility_coupling_2,)
            hgf.attributes[2]["volatility_coupling_children"] = (volatility_coupling_2,)

    surprise = hgf.input_data(input_data=input_data, time_steps=time_steps).surprise(
        response_function=response_function,
        response_function_inputs=response_function_inputs,
        response_function_parameters=response_function_parameters,
    )

    return -surprise


def hgf_logp(
    mean_1: ArrayLike = 0.0,
    mean_2: ArrayLike = 0.0,
    mean_3: ArrayLike = 0.0,
    precision_1: ArrayLike = 1.0,
    precision_2: ArrayLike = 1.0,
    precision_3: ArrayLike = 1.0,
    tonic_volatility_1: ArrayLike = -3.0,
    tonic_volatility_2: ArrayLike = -3.0,
    tonic_volatility_3: ArrayLike = -3.0,
    tonic_drift_1: ArrayLike = 0.0,
    tonic_drift_2: ArrayLike = 0.0,
    tonic_drift_3: ArrayLike = 0.0,
    volatility_coupling_1: ArrayLike = 1.0,
    volatility_coupling_2: ArrayLike = 1.0,
    input_precision: ArrayLike = np.inf,
    response_function_parameters: ArrayLike = np.ones(1),
    vectorized_logp: Callable = logp,
    input_data: ArrayLike = np.nan,
    response_function_inputs: ArrayLike = np.nan,
    time_steps: ArrayLike = np.nan,
) -> Array:
    """Compute log-probabilities of a batch of Hierarchical Gaussian Filters.

    .. hint::

       This function supports broadcasting along the first axis, which means that it can
       fit multiple models when input data are provided. When a network parameter is a
       float, this value will be used on all models. When a network parameter is an
       array, the size should match the number of input data, and different values will
       be used accordingly.

    Parameters
    ----------
    mean_1 :
        The mean at the first level of the HGF.
    mean_2 :
        The mean at the second level of the HGF.
    mean_3 :
        The mean at the third level of the HGF. The value of this parameter will be
        ignored when using a two-level HGF (`n_levels=2`).
    precision_1 :
        The precision at the first level of the HGF.
    precision_2 :
        The precision at the second level of the HGF.
    precision_3 :
        The precision at the third level of the HGF. The value of this parameter will
        be ignored when using a two-level HGF (`n_levels=2`).
    tonic_volatility_1 :
        The tonic volatility at the first level of the HGF. This parameter represents
        the tonic part of the variance (the part that is not inherited from parent
        nodes).
    tonic_volatility_2 :
        The tonic volatility at the second level of the HGF. This parameter represents
        the tonic part of the variance (the part that is not inherited from parent
        nodes).
    tonic_volatility_3 :
        The tonic volatility at the third level of the HGF. This parameter represents
        the tonic part of the variance (the part that is not inherited from parent
        nodes). The value of this parameter will be ignored when using a two-level HGF
        (`n_levels=2`).
    tonic_drift_1 :
        The tonic drift at the first level of the HGF. This parameter represents the
        drift of the random walk.
    tonic_drift_2 :
        The tonic drift at the second level of the HGF. This parameter represents the
        drift of the random walk.
    tonic_drift_3 :
        The tonic drift at the first level of the HGF. This parameter represents the
        drift of the random walk. The value of this parameter will be ignored when
        using a two-level HGF (`n_levels=2`).
    volatility_coupling_1 :
        The volatility coupling between the first and second levels of the HGF. This
        represents the phasic part of the variance (the part affected by the
        parent nodes). Defaults to `1.0`.
    volatility_coupling_2 :
        The volatility coupling between the second and third levels of the HGF. This
        represents the phasic part of the variance (the part affected by the
        parent nodes). Defaults to `1.0`. The value of this parameter will be
        ignored when using a two-level HGF (`n_levels=2`).
    input_precision :
        The expected precision associated with the continuous or binary input, depending
        on the model type. The default is `np.inf`.
    response_function_parameters :
        An array list of additional parameters that will be passed to the response
        function. This can include values over which inference is performed in a PyMC
        model (e.g. the inverse temperature of a binary softmax).
    vectorized_logp :
        A vectorized log probability function for a two or three-layered HGF.
    input_data :
        An array of input time series where the first dimension is the number of models
        to fit in parallel.
    response_function_inputs :
        An array of behavioural input passed to the response function where the first
        dimension is the number of models to fit in parallel.
    time_steps :
        An array of input time steps where the first dimension is the number of models
        to fit in parallel.

    Returns
    -------
    log_prob :
        The sum of the log probabilities (negative surprise).

    """
    # number of models
    n = input_data.shape[0]

    # ensure that the response parameters have n_model as first dimension
    response_function_parameters = jnp.broadcast_to(
        response_function_parameters, (n,) + response_function_parameters.shape[1:]
    )

    # Broadcast inputs to an array with length n>=1
    (
        _tonic_volatility_1,
        _tonic_volatility_2,
        _tonic_volatility_3,
        _input_precision,
        _tonic_drift_1,
        _tonic_drift_2,
        _tonic_drift_3,
        _precision_1,
        _precision_2,
        _precision_3,
        _mean_1,
        _mean_2,
        _mean_3,
        _volatility_coupling_1,
        _volatility_coupling_2,
        _,
    ) = jnp.broadcast_arrays(
        tonic_volatility_1,
        tonic_volatility_2,
        tonic_volatility_3,
        input_precision,
        tonic_drift_1,
        tonic_drift_2,
        tonic_drift_3,
        precision_1,
        precision_2,
        precision_3,
        mean_1,
        mean_2,
        mean_3,
        volatility_coupling_1,
        volatility_coupling_2,
        jnp.zeros(n),
    )

    logp = vectorized_logp(
        input_data=input_data,
        response_function_inputs=response_function_inputs,
        response_function_parameters=response_function_parameters,
        time_steps=time_steps,
        mean_1=_mean_1,
        mean_2=_mean_2,
        mean_3=_mean_3,
        precision_1=_precision_1,
        precision_2=_precision_2,
        precision_3=_precision_3,
        tonic_volatility_1=_tonic_volatility_1,
        tonic_volatility_2=_tonic_volatility_2,
        tonic_volatility_3=_tonic_volatility_3,
        tonic_drift_1=_tonic_drift_1,
        tonic_drift_2=_tonic_drift_2,
        tonic_drift_3=_tonic_drift_3,
        volatility_coupling_1=_volatility_coupling_1,
        volatility_coupling_2=_volatility_coupling_2,
        input_precision=_input_precision,
    )

    # Return the log probabilities (negative surprise)
    return logp.sum(), logp


class HGFLogpGradOp(Op):
    """Gradient Op for the HGF distribution."""

    def __init__(
        self,
        input_data: Union[ArrayLike] = np.nan,
        time_steps: Optional[ArrayLike] = None,
        model_type: str = "continous",
        n_levels: int = 2,
        response_function: Optional[Callable] = None,
        response_function_inputs: Optional[ArrayLike] = None,
    ):
        """Initialize function.

        Parameters
        ----------
        input_data :
            An array of input time series where the first dimension is the number of
            models to fit in parallel. By default, the associated time_steps vector is
            the unit vector. A different time vector can be passed to the `time_steps`
            argument.
        time_steps :
            An array with shapes equal to input_data containing the time_steps vectors.
            If one of the list items is `None`, or if `None` is provided instead, the
            time_steps vector will default to an integer vector starting at 0.
        model_type :
            The model type to use (can be "continuous" or "binary").
        n_levels :
            The number of hierarchies in the perceptual model (can be `2` or `3`). If
            `None`, the nodes hierarchy is not created and might be provided afterwards
            using `add_nodes()`.
        response_function :
            The response function to use to compute the model surprise.
        response_function_inputs :
            A list of tuples with the same length as the number of models. Each tuple
            contains additional data and parameters that can be accessible to the
            response functions.

        """
        if time_steps is None:
            time_steps = np.ones(shape=input_data.shape)

        # create the default HGF template to be use by the logp function
        self.hgf = HGF(n_levels=n_levels, model_type=model_type)

        # create a vectorized version of the logp function
        vectorized_logp = vmap(
            Partial(
                logp,
                hgf=self.hgf,
                response_function=response_function,
            )
        )

        # define the gradient of the vectorized function
        self.grad_logp = jit(
            grad(
                Partial(
                    hgf_logp,
                    vectorized_logp=vectorized_logp,
                    input_data=input_data,
                    time_steps=time_steps,
                    response_function_inputs=response_function_inputs,
                ),
                argnums=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15],
                has_aux=True,
            )
        )

    def make_node(
        self,
        mean_1: ArrayLike = np.array(0.0),
        mean_2: ArrayLike = np.array(0.0),
        mean_3: ArrayLike = np.array(0.0),
        precision_1: ArrayLike = np.array(1.0),
        precision_2: ArrayLike = np.array(1.0),
        precision_3: ArrayLike = np.array(1.0),
        tonic_volatility_1: ArrayLike = np.array(-3.0),
        tonic_volatility_2: ArrayLike = np.array(-3.0),
        tonic_volatility_3: ArrayLike = np.array(-3.0),
        tonic_drift_1: ArrayLike = np.array(0.0),
        tonic_drift_2: ArrayLike = np.array(0.0),
        tonic_drift_3: ArrayLike = np.array(0.0),
        volatility_coupling_1: ArrayLike = np.array(1.0),
        volatility_coupling_2: ArrayLike = np.array(1.0),
        input_precision: ArrayLike = np.array(1e4),
        response_function_parameters: ArrayLike = np.array(1.0),
    ):
        """Initialize node structure."""
        # Convert our inputs to symbolic variables
        inputs = [
            pt.as_tensor_variable(mean_1, ndim=1),
            pt.as_tensor_variable(mean_2, ndim=1),
            pt.as_tensor_variable(mean_3, ndim=1),
            pt.as_tensor_variable(precision_1, ndim=1),
            pt.as_tensor_variable(precision_2, ndim=1),
            pt.as_tensor_variable(precision_3, ndim=1),
            pt.as_tensor_variable(tonic_volatility_1, ndim=1),
            pt.as_tensor_variable(tonic_volatility_2, ndim=1),
            pt.as_tensor_variable(tonic_volatility_3, ndim=1),
            pt.as_tensor_variable(tonic_drift_1, ndim=1),
            pt.as_tensor_variable(tonic_drift_2, ndim=1),
            pt.as_tensor_variable(tonic_drift_3, ndim=1),
            pt.as_tensor_variable(volatility_coupling_1, ndim=1),
            pt.as_tensor_variable(volatility_coupling_2, ndim=1),
            pt.as_tensor_variable(input_precision, ndim=1),
            pt.as_tensor_variable(response_function_parameters),
        ]
        # This `Op` will return one gradient per input. For simplicity, we assume
        # each output is of the same type as the input. In practice, you should use
        # the exact dtype to avoid overhead when saving the results of the computation
        # in `perform`
        outputs = [inp.type() for inp in inputs]
        return Apply(self, inputs, outputs)

    def perform(
        self, node, inputs: List[TensorVariable], outputs: List[TensorVariable]
    ):
        """Perform node operations."""
        (
            grad_mean_1,
            grad_mean_2,
            grad_mean_3,
            grad_precision_1,
            grad_precision_2,
            grad_precision_3,
            grad_tonic_volatility_1,
            grad_tonic_volatility_2,
            grad_tonic_volatility_3,
            grad_tonic_drift_1,
            grad_tonic_drift_2,
            grad_tonic_drift_3,
            grad_volatility_coupling_1,
            grad_volatility_coupling_2,
            grad_input_precision,
            grad_response_function_parameters,
        ), _ = self.grad_logp(*inputs)

        outputs[0][0] = np.asarray(grad_mean_1, dtype=node.outputs[0].dtype)
        outputs[1][0] = np.asarray(grad_mean_2, dtype=node.outputs[1].dtype)
        outputs[2][0] = np.asarray(grad_mean_3, dtype=node.outputs[2].dtype)
        outputs[3][0] = np.asarray(grad_precision_1, dtype=node.outputs[3].dtype)
        outputs[4][0] = np.asarray(grad_precision_2, dtype=node.outputs[4].dtype)
        outputs[5][0] = np.asarray(grad_precision_3, dtype=node.outputs[5].dtype)
        outputs[6][0] = np.asarray(grad_tonic_volatility_1, dtype=node.outputs[6].dtype)
        outputs[7][0] = np.asarray(grad_tonic_volatility_2, dtype=node.outputs[7].dtype)
        outputs[8][0] = np.asarray(grad_tonic_volatility_3, dtype=node.outputs[8].dtype)
        outputs[9][0] = np.asarray(grad_tonic_drift_1, dtype=node.outputs[9].dtype)
        outputs[10][0] = np.asarray(grad_tonic_drift_2, dtype=node.outputs[10].dtype)
        outputs[11][0] = np.asarray(grad_tonic_drift_3, dtype=node.outputs[11].dtype)
        outputs[12][0] = np.asarray(
            grad_volatility_coupling_1, dtype=node.outputs[12].dtype
        )
        outputs[13][0] = np.asarray(
            grad_volatility_coupling_2, dtype=node.outputs[13].dtype
        )
        outputs[14][0] = np.asarray(grad_input_precision, dtype=node.outputs[14].dtype)
        outputs[15][0] = np.asarray(
            grad_response_function_parameters, dtype=node.outputs[15].dtype
        )


class HGFDistribution(Op):
    r"""The HGF distribution PyMC >= 5.0 compatible.

    Examples
    --------
    This class should be used in the context of a PyMC probabilistic model to estimate
    one or many parameter-s probability density with MCMC sampling.

    .. code-block:: python

        import arviz as az
        import numpy as np
        import pymc as pm
        from math import log
        from pyhgf import load_data
        from pyhgf.distribution import HGFDistribution
        from pyhgf.response import total_gaussian_surprise

    Create the data (value and time vectors)

    .. code-block:: python

        timeserie = load_data("continuous")

    We create the PyMC distribution here, specifying the type of model and response
    function we want to use (i.e. simple Gaussian surprise).

    .. code-block:: python

        hgf_logp_op = HGFDistribution(
            n_levels=2,
            input_data=timeserie[np.newaxis],
            response_function=total_gaussian_surprise,
        )

    Create the PyMC model. Here we are fixing all the parameters to arbitrary values and
    sampling the posterior probability density of the tonic volatility at the second
    level, using a normal distribution as prior:

    .. math::

        \omega_2 \sim \mathcal{N}(-11.0, 2.0)

    .. code-block:: python

        with pm.Model() as model:

            # prior over tonic volatility
            ω_2 = pm.Normal("ω_2", -11.0, 2)

            pm.Potential(
                "hhgf_loglike",
                hgf_logp_op(
                    tonic_volatility_2=ω_2,
                    precision_1=1.0,
                    precision_2=1.0,
                    mean_1=1.0,
                ),
            )

    Sample the model - Using 2 chains, 1 cores and 1000 warmups.

    .. code-block:: python

        with model:
            idata = pm.sample(chains=2, cores=1, tune=1000)

    Print a summary of the results using Arviz

    >>> az.summary(idata, var_names="ω_2")

    ===  =======  =====  =======  =======  =========  =======  ========  ========  =====
    ..      mean     sd   hdi_3%  hdi_97%  mcse_mean  mcse_sd  ess_bulk  ess_tail  r_hat
    ===  =======  =====  =======  =======  =========  =======  ========  ========  =====
    ω_2  -14.776  1.065  -16.766  -12.944      0.038    0.027       806       938      1
    ===  =======  =====  =======  =======  =========  =======  ========  ========  =====

    """

    def __init__(
        self,
        input_data: ArrayLike = jnp.nan,
        time_steps: Optional[ArrayLike] = None,
        model_type: str = "continuous",
        n_levels: int = 2,
        response_function: Optional[Callable] = None,
        response_function_inputs: Optional[ArrayLike] = None,
    ):
        """Distribution initialization.

        Parameters
        ----------
        input_data :
            List of input data. When `n` models should be fitted, the list contains `n`
            1d Numpy arrays. By default, the associated time vector is the unit
            vector starting at `0`. A different time_steps vector can be passed to
            the `time_steps` argument.
        time_steps :
            List of 1d Numpy arrays containing the time_steps vectors for each input
            time series. If one of the list items is `None`, or if `None` is provided
            instead, the time vector will default to an integers vector starting at 0.
        model_type :
            The model type to use (can be "continuous" or "binary").
        n_levels :
            The number of hierarchies in the perceptual model (can be `2` or `3`). If
            `None`, the nodes hierarchy is not created and might be provided afterwards
            using `add_nodes()`.
        response_function :
            The response function to use to compute the model surprise.
        response_function_inputs :
            A list of tuples with the same length as the number of models. Each tuple
            contains additional data and parameters that can be accessible to the
            response functions.

        """
        if time_steps is None:
            time_steps = np.ones(shape=input_data.shape)

        assert time_steps.shape == input_data.shape

        # create the default HGF template to be use by the logp function
        self.hgf = HGF(n_levels=n_levels, model_type=model_type)

        # create a vectorized version of the logp function
        vectorized_logp = vmap(
            Partial(
                logp,
                hgf=self.hgf,
                response_function=response_function,
            )
        )

        # The value function
        self.hgf_logp = jit(
            Partial(
                hgf_logp,
                vectorized_logp=vectorized_logp,
                input_data=input_data,
                time_steps=time_steps,
                response_function_inputs=response_function_inputs,
            ),
        )

        # The gradient function
        self.hgf_logp_grad_op = HGFLogpGradOp(
            input_data=input_data,
            time_steps=time_steps,
            model_type=model_type,
            n_levels=n_levels,
            response_function=response_function,
            response_function_inputs=response_function_inputs,
        )

    def make_node(
        self,
        mean_1: ArrayLike = np.array(0.0),
        mean_2: ArrayLike = np.array(0.0),
        mean_3: ArrayLike = np.array(0.0),
        precision_1: ArrayLike = np.array(1.0),
        precision_2: ArrayLike = np.array(1.0),
        precision_3: ArrayLike = np.array(1.0),
        tonic_volatility_1: ArrayLike = np.array(-3.0),
        tonic_volatility_2: ArrayLike = np.array(-3.0),
        tonic_volatility_3: ArrayLike = np.array(-3.0),
        tonic_drift_1: ArrayLike = np.array(0.0),
        tonic_drift_2: ArrayLike = np.array(0.0),
        tonic_drift_3: ArrayLike = np.array(0.0),
        volatility_coupling_1: ArrayLike = np.array(1.0),
        volatility_coupling_2: ArrayLike = np.array(1.0),
        input_precision: ArrayLike = np.array(1e4),
        response_function_parameters: ArrayLike = np.array(1.0),
    ):
        """Convert inputs to symbolic variables."""
        inputs = [
            pt.as_tensor_variable(mean_1, ndim=1),
            pt.as_tensor_variable(mean_2, ndim=1),
            pt.as_tensor_variable(mean_3, ndim=1),
            pt.as_tensor_variable(precision_1, ndim=1),
            pt.as_tensor_variable(precision_2, ndim=1),
            pt.as_tensor_variable(precision_3, ndim=1),
            pt.as_tensor_variable(tonic_volatility_1, ndim=1),
            pt.as_tensor_variable(tonic_volatility_2, ndim=1),
            pt.as_tensor_variable(tonic_volatility_3, ndim=1),
            pt.as_tensor_variable(tonic_drift_1, ndim=1),
            pt.as_tensor_variable(tonic_drift_2, ndim=1),
            pt.as_tensor_variable(tonic_drift_3, ndim=1),
            pt.as_tensor_variable(volatility_coupling_1, ndim=1),
            pt.as_tensor_variable(volatility_coupling_2, ndim=1),
            pt.as_tensor_variable(input_precision, ndim=1),
            pt.as_tensor_variable(response_function_parameters),
        ]
        # Define the type of output returned by the wrapped JAX function
        outputs = [pt.scalar(dtype=float)]
        return Apply(self, inputs, outputs)

    def perform(
        self, node, inputs: List[TensorVariable], outputs: List[TensorVariable]
    ):
        """Run the function forward."""
        log_likelihood, _ = self.hgf_logp(*inputs)
        outputs[0][0] = np.asarray(log_likelihood, dtype=float)

    def grad(
        self, inputs: List[TensorVariable], output_gradients: List[TensorVariable]
    ) -> List[TensorVariable]:
        """Gradient of the function."""
        (
            grad_mean_1,
            grad_mean_2,
            grad_mean_3,
            grad_precision_1,
            grad_precision_2,
            grad_precision_3,
            grad_tonic_volatility_1,
            grad_tonic_volatility_2,
            grad_tonic_volatility_3,
            grad_tonic_drift_1,
            grad_tonic_drift_2,
            grad_tonic_drift_3,
            grad_volatility_coupling_1,
            grad_volatility_coupling_2,
            grad_input_precision,
            grad_response_function_parameters,
        ) = self.hgf_logp_grad_op(*inputs)

        # If there are inputs for which the gradients will never be needed or cannot
        # be computed, `pytensor.gradient.grad_not_implemented` should  be used as the
        # output gradient for that input.
        output_gradient = output_gradients[0]
        return [
            output_gradient * grad_mean_1,
            output_gradient * grad_mean_2,
            output_gradient * grad_mean_3,
            output_gradient * grad_precision_1,
            output_gradient * grad_precision_2,
            output_gradient * grad_precision_3,
            output_gradient * grad_tonic_volatility_1,
            output_gradient * grad_tonic_volatility_2,
            output_gradient * grad_tonic_volatility_3,
            output_gradient * grad_tonic_drift_1,
            output_gradient * grad_tonic_drift_2,
            output_gradient * grad_tonic_drift_3,
            output_gradient * grad_volatility_coupling_1,
            output_gradient * grad_volatility_coupling_2,
            output_gradient * grad_input_precision,
            output_gradient * grad_response_function_parameters,
        ]


class HGFPointwise(Op):
    """The HGF distribution returning pointwise log probability.

    This class should be used in the context of model comparison where the pointwise log
    probabilities are required for cross-validation.

    """

    def __init__(
        self,
        input_data: ArrayLike = jnp.nan,
        time_steps: Optional[ArrayLike] = None,
        model_type: str = "continuous",
        n_levels: int = 2,
        response_function: Optional[Callable] = None,
        response_function_inputs: Optional[ArrayLike] = None,
    ):
        """Distribution initialization.

        Parameters
        ----------
        input_data :
            List of input data. When `n` models should be fitted, the list contains `n`
            1d Numpy arrays. By default, the associated time vector is the unit
            vector starting at `0`. A different time_steps vector can be passed to
            the `time_steps` argument.
        time_steps :
            List of 1d Numpy arrays containing the time_steps vectors for each input
            time series. If one of the list items is `None`, or if `None` is provided
            instead, the time vector will default to an integers vector starting at 0.
        model_type :
            The model type to use (can be "continuous" or "binary").
        n_levels :
            The number of hierarchies in the perceptual model (can be `2` or `3`). If
            `None`, the nodes hierarchy is not created and might be provided afterwards
            using `add_nodes()`.
        response_function :
            The response function to use to compute the model surprise.
        response_function_inputs :
            A list of tuples with the same length as the number of models. Each tuple
            contains additional data and parameters that can be accessible to the
            response functions.

        """
        if time_steps is None:
            time_steps = np.ones(shape=input_data.shape)

        # create the default HGF template to be use by the logp function
        self.hgf = HGF(n_levels=n_levels, model_type=model_type)

        # create a vectorized version of the logp function
        vectorized_logp = vmap(
            Partial(
                logp,
                hgf=self.hgf,
                response_function=response_function,
            )
        )

        # The value function
        self.hgf_logp = jit(
            Partial(
                hgf_logp,
                vectorized_logp=vectorized_logp,
                input_data=input_data,
                time_steps=time_steps,
                response_function_inputs=response_function_inputs,
            ),
        )

    def make_node(
        self,
        mean_1: ArrayLike = np.array(0.0),
        mean_2: ArrayLike = np.array(0.0),
        mean_3: ArrayLike = np.array(0.0),
        precision_1: ArrayLike = np.array(1.0),
        precision_2: ArrayLike = np.array(1.0),
        precision_3: ArrayLike = np.array(1.0),
        tonic_volatility_1: ArrayLike = np.array(-3.0),
        tonic_volatility_2: ArrayLike = np.array(-3.0),
        tonic_volatility_3: ArrayLike = np.array(-3.0),
        tonic_drift_1: ArrayLike = np.array(0.0),
        tonic_drift_2: ArrayLike = np.array(0.0),
        tonic_drift_3: ArrayLike = np.array(0.0),
        volatility_coupling_1: ArrayLike = np.array(1.0),
        volatility_coupling_2: ArrayLike = np.array(1.0),
        input_precision: ArrayLike = np.array(1e4),
        response_function_parameters: ArrayLike = np.array(1.0),
    ):
        """Convert inputs to symbolic variables."""
        inputs = [
            pt.as_tensor_variable(mean_1),
            pt.as_tensor_variable(mean_2),
            pt.as_tensor_variable(mean_3),
            pt.as_tensor_variable(precision_1),
            pt.as_tensor_variable(precision_2),
            pt.as_tensor_variable(precision_3),
            pt.as_tensor_variable(tonic_volatility_1),
            pt.as_tensor_variable(tonic_volatility_2),
            pt.as_tensor_variable(tonic_volatility_3),
            pt.as_tensor_variable(tonic_drift_1),
            pt.as_tensor_variable(tonic_drift_2),
            pt.as_tensor_variable(tonic_drift_3),
            pt.as_tensor_variable(volatility_coupling_1),
            pt.as_tensor_variable(volatility_coupling_2),
            pt.as_tensor_variable(input_precision),
            pt.as_tensor_variable(response_function_parameters),
        ]
        # Define the type of output returned by the wrapped JAX function
        outputs = [pt.matrix(dtype=float)]
        return Apply(self, inputs, outputs)

    def perform(
        self, node, inputs: List[TensorVariable], outputs: List[TensorVariable]
    ):
        """Run the function forward."""
        _, pointwise_log_likelihood = self.hgf_logp(*inputs)
        outputs[0][0] = np.asarray(pointwise_log_likelihood, dtype=float)



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from functools import partial
from typing import Dict

from jax import jit

from pyhgf.typing import Edges


@partial(jit, static_argnames=("node_idx", "edges"))
def generic_state_prediction_error(
    attributes: Dict, edges: Edges, node_idx: int, **args
) -> Dict:
    """Prediction error from a generic input node.

    A generic input node receive an observation and pass it as it is to its value
    parents.

    Parameters
    ----------
    attributes :
        The attributes of the probabilistic nodes.
    edges :
        The edges of the probabilistic nodes as a tuple of
        :py:class:`pyhgf.typing.Indexes`. The tuple has the same length as the node
        number. For each node, the index lists the value and volatility parents and
        children.
    node_idx :
        Pointer to the value parent node that will be updated.

    Returns
    -------
    attributes :
        The attributes of the probabilistic nodes.

    """
    for value_parent_idx in edges[node_idx].value_parents:  # type: ignore
        attributes[value_parent_idx]["mean"] = attributes[node_idx]["mean"]

    return attributes



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from copy import deepcopy
from typing import Callable, Dict, List, Optional, Tuple, Union

import jax.numpy as jnp
import numpy as np
import pandas as pd
from jax.lax import scan, switch
from jax.tree_util import Partial
from jax.typing import ArrayLike

from pyhgf.plots import plot_correlations, plot_network, plot_nodes, plot_trajectories
from pyhgf.typing import (
    AdjacencyLists,
    Attributes,
    Edges,
    NetworkParameters,
    UpdateSequence,
)
from pyhgf.utils import (
    add_edges,
    beliefs_propagation,
    fill_categorical_state_node,
    get_input_idxs,
    get_update_sequence,
    to_pandas,
)


class Network:
    """A predictive coding neural network.

    This is the core class to define and manipulate neural networks, that consists in
    1. attributes, 2. structure and 3. update sequences.

    Attributes
    ----------
    attributes :
        The attributes of the probabilistic nodes.
    edges :
        The edges of the probabilistic nodes as a tuple of
        :py:class:`pyhgf.typing.AdjacencyLists`. The tuple has the same length as the
        node number. For each node, the index lists the value/volatility
        parents/children.
    inputs :
        Information on the input nodes.
    node_trajectories :
        The dynamic of the node's beliefs after updating.
    update_sequence :
        The sequence of update functions that are applied during the belief propagation
        step.
    scan_fn :
        The function that is passed to :py:func:`jax.lax.scan`. This is a pre-
        parametrized version of :py:func:`pyhgf.networks.beliefs_propagation`.

    """

    def __init__(self) -> None:
        """Initialize an empty neural network."""
        self.edges: Edges = ()
        self.node_trajectories: Dict = {}
        self.attributes: Attributes = {-1: {"time_step": 0.0}}
        self.update_sequence: Optional[UpdateSequence] = None
        self.scan_fn: Optional[Callable] = None

    @property
    def input_idxs(self):
        """Idexes of state nodes that can observe new data points by default."""
        input_idxs = get_input_idxs(self.edges)

        # set the autoconnection strength and tonic volatility to 0
        for idx in input_idxs:
            if self.edges[idx].node_type == 2:
                self.attributes[idx]["autoconnection_strength"] = 0.0
                self.attributes[idx]["tonic_volatility"] = 0.0
        return input_idxs

    @input_idxs.setter
    def input_idxs(self, value):
        self.input_idxs = value

    def create_belief_propagation_fn(
        self, overwrite: bool = True, update_type: str = "eHGF"
    ) -> "Network":
        """Create the belief propagation function.

        .. note:
           This step is called by default when using py:meth:`input_data`.

        Parameters
        ----------
        overwrite :
            If `True` (default), create a new belief propagation function and ignore
            preexisting values. Otherwise, do not create a new function if the attribute
            `scan_fn` is already defined.
        update_type :
            The type of update to perform for volatility coupling. Can be `"eHGF"`
            (defaults) or `"standard"`. The eHGF update step was proposed as an
            alternative to the original definition in that it starts by updating the
            mean and then the precision of the parent node, which generally reduces the
            errors associated with impossible parameter space and improves sampling.

        """
        # create the update sequence if it does not already exist
        if self.update_sequence is None:
            self.update_sequence = get_update_sequence(
                network=self, update_type=update_type
            )

        # create the belief propagation function
        # this function is used by scan to loop over observations
        if (self.scan_fn is None) or overwrite:
            self.scan_fn = Partial(
                beliefs_propagation,
                update_sequence=self.update_sequence,
                edges=self.edges,
                input_idxs=self.input_idxs,
            )

        return self

    def input_data(
        self,
        input_data: Union[np.ndarray, tuple],
        time_steps: Optional[np.ndarray] = None,
        observed: Optional[Union[np.ndarray, tuple]] = None,
        input_idxs: Optional[Tuple[int]] = None,
    ):
        """Add new observations.

        Parameters
        ----------
        input_data :
            2d array of new observations (time x features).
        time_steps :
            Time steps vector (optional). If `None`, this will default to
            `np.ones(len(input_data))`.
        observed :
            A time * input node boolean array masking `input_data`. In case of missing
            inputs, (i.e. `observed` is `0`), the input node will have value and
            volatility set to `0.0`. If the parent(s) of this input receive prediction
            error from other children, they simply ignore this one. If they are not
            receiving other prediction errors, they are updated by keeping the same
            mean by decreasing the precision as a function of time to reflect the
            evolution of the underlying Gaussian Random Walk.
        .. warning::
            Missing inputs are missing observations from the agent's perspective and
            should not be used to handle missing data points that were observed (e.g.
            missing in the event log, or rejected trials).
        input_idxs :
            Indexes on the state nodes receiving observations.

        """
        # set the input nodes indexes
        if input_idxs is not None:
            self.input_idxs = input_idxs

        # belief propagation function
        if self.scan_fn is None:
            self = self.create_belief_propagation_fn()

        # input_data should be a tuple of n by time_steps arrays
        if not isinstance(input_data, tuple):
            if observed is None:
                observed = np.ones(input_data.shape, dtype=int)
            if input_data.ndim == 1:

                # Interleave observations and masks
                input_data = (input_data, observed)
            else:
                observed = jnp.hsplit(observed, input_data.shape[1])
                observations = jnp.hsplit(input_data, input_data.shape[1])

                # Interleave observations and masks
                input_data = tuple(
                    [
                        item.flatten()
                        for pair in zip(observations, observed)
                        for item in pair
                    ]
                )

        # time steps vector
        if time_steps is None:
            time_steps = np.ones(input_data[0].shape[0])

        # this is where the model loops over the whole input time series
        # at each time point, the node structure is traversed and beliefs are updated
        # using precision-weighted prediction errors
        last_attributes, node_trajectories = scan(
            self.scan_fn, self.attributes, (*input_data, time_steps)
        )

        # belief trajectories
        self.node_trajectories = node_trajectories
        self.last_attributes = last_attributes

        return self

    def input_custom_sequence(
        self,
        update_branches: Tuple[UpdateSequence],
        branches_idx: np.array,
        input_data: np.ndarray,
        time_steps: Optional[np.ndarray] = None,
        observed: Optional[np.ndarray] = None,
        input_idxs: Optional[Tuple[int]] = None,
    ):
        """Add new observations with custom update sequences.

        This method should be used when the update sequence is function of the input
        data. (e.g. in the case of missing/null observations that should not trigger
        node update).

        .. note::
           When the dynamic adaptation of the update sequence is not required, it is
           recommended to use :py:meth:`pyhgf.model.HGF.input_data` instead as this
           might result in performance improvement.

        Parameters
        ----------
        update_branches :
            A tuple of UpdateSequence listing the possible update sequences.
        branches_idx :
            The branches indexes (integers). Should have the same length as the input
            data.
        input_data :
            2d array of new observations (time x features).
        time_steps :
            Time vector (optional). If `None`, the time vector will default to
            `np.ones(len(input_data))`. This vector is automatically transformed
            into a time steps vector.
        observed :
            A 2d boolean array masking `input_data`. In case of missing inputs, (i.e.
            `observed` is `0`), the input node will have value and volatility set to
            `0.0`. If the parent(s) of this input receive prediction error from other
            children, they simply ignore this one. If they are not receiving other
            prediction errors, they are updated by keeping the same mean be decreasing
            the precision as a function of time to reflect the evolution of the
            underlying Gaussian Random Walk.
        .. warning::
            Missing inputs are missing observations from the agent's perspective and
            should not be used to handle missing data points that were observed (e.g.
            missing in the event log, or rejected trials).
        input_idxs :
            Indexes on the state nodes receiving observations.

        """
        # set the input nodes indexes
        if input_idxs is not None:
            self.input_idxs = input_idxs

        # input_data should be a tuple of n by time_steps arrays
        if not isinstance(input_data, tuple):
            if observed is None:
                observed = np.ones(input_data.shape, dtype=int)
            if input_data.ndim == 1:

                # Interleave observations and masks
                input_data = (input_data, observed)
            else:
                observed = jnp.hsplit(observed, input_data.shape[1])
                observations = jnp.hsplit(input_data, input_data.shape[1])

                # Interleave observations and masks
                input_data = tuple(
                    [item for pair in zip(observations, observed) for item in pair]
                )

        # time steps vector
        if time_steps is None:
            time_steps = np.ones(input_data[0].shape[0])

        # create the update functions that will be scanned
        branches_fn = [
            Partial(
                beliefs_propagation,
                update_sequence=seq,
                edges=self.edges,
                input_idxs=self.input_idxs,
            )
            for seq in update_branches
        ]

        # create the function that will be scanned
        def switching_propagation(attributes, scan_input):
            (*data, idx) = scan_input
            return switch(idx, branches_fn, attributes, data)

        # wrap the inputs
        scan_input = (*input_data, time_steps, branches_idx)

        # scan over the input data and apply the switching belief propagation functions
        _, node_trajectories = scan(switching_propagation, self.attributes, scan_input)

        # the node structure at each value updates
        self.node_trajectories = node_trajectories

        return self

    def get_network(self) -> NetworkParameters:
        """Return the attributes, edges and update sequence defining the network."""
        if self.scan_fn is None:
            self = self.create_belief_propagation_fn()

        assert self.update_sequence is not None

        return self.attributes, self.edges, self.update_sequence

    def add_nodes(
        self,
        kind: str = "continuous-state",
        n_nodes: int = 1,
        node_parameters: Dict = {},
        value_children: Optional[Union[List, Tuple, int]] = None,
        value_parents: Optional[Union[List, Tuple, int]] = None,
        volatility_children: Optional[Union[List, Tuple, int]] = None,
        volatility_parents: Optional[Union[List, Tuple, int]] = None,
        coupling_fn: Tuple[Optional[Callable], ...] = (None,),
        **additional_parameters,
    ):
        """Add new input/state node(s) to the neural network.

        Parameters
        ----------
        kind :
            The kind of node to create. If `"continuous-state"` (default), the node will
            be a regular state node that can have value and/or volatility
            parents/children. If `"binary-state"`, the node should be the
            value parent of a binary input. State nodes filtering distribution from the
            exponential family can be created using the `"ef-"` prefix (e.g.
            `"ef-normal"` for a univariate normal distribution). Note that only a few
            distributions are implemented at the moment.

            In addition to state nodes, four types of input nodes are supported:
            - `generic-input`: receive a value or an array and pass it to the parent
            nodes.
            - `continuous-input`: receive a continuous observation as input.
            - `binary-input` receives a single boolean as observation. The parameters
            provided to the binary input node contain: 1. `binary_precision`, the binary
            input precision, which defaults to `jnp.inf`. 2. `eta0`, the lower bound of
            the binary process, which defaults to `0.0`. 3. `eta1`, the higher bound of
            the binary process, which defaults to `1.0`.
            - `categorical-input` receives a boolean array as observation. The
            parameters provided to the categorical input node contain: 1.
            `n_categories`, the number of categories implied by the categorical state.

        .. note::
            When using a categorical state node, the `binary_parameters` can be used to
            parametrize the implied collection of binary HGFs.

        .. note:
            When using `categorical-input`, the implied `n` binary HGFs are
            automatically created with a shared volatility parent at the third level,
            resulting in a network with `3n + 2` nodes in total.

        n_nodes :
            The number of nodes to create (defaults to `1`).
        node_parameters :
            Dictionary of parameters. The default values are automatically inferred
            from the node type. Different values can be provided by passing them in the
            dictionary, which will overwrite the defaults.
        value_children :
            Indexes to the node's value children. The index can be passed as an integer
            or a list of integers, in case of multiple children. The coupling strength
            can be controlled by passing a tuple, where the first item is the list of
            indexes, and the second item is the list of coupling strengths.
        value_parents :
            Indexes to the node's value parents. The index can be passed as an integer
            or a list of integers, in case of multiple children. The coupling strength
            can be controlled by passing a tuple, where the first item is the list of
            indexes, and the second item is the list of coupling strengths.
        volatility_children :
            Indexes to the node's volatility children. The index can be passed as an
            integer or a list of integers, in case of multiple children. The coupling
            strength can be controlled by passing a tuple, where the first item is the
            list of indexes, and the second item is the list of coupling strengths.
        volatility_parents :
            Indexes to the node's volatility parents. The index can be passed as an
            integer or a list of integers, in case of multiple children. The coupling
            strength can be controlled by passing a tuple, where the first item is the
            list of indexes, and the second item is the list of coupling strengths.
        coupling_fn :
            Coupling function(s) between the current node and its value children.
            It has to be provided as a tuple. If multiple value children are specified,
            the coupling functions must be stated in the same order of the children.
            Note: if a node has multiple parents nodes with different coupling
            functions, a coupling function should be indicated for all the parent nodes.
            If no coupling function is stated, the relationship between nodes is assumed
            linear.
        **kwargs :
            Additional keyword parameters will be passed and overwrite the node
            attributes.

        """
        if kind not in [
            "DP-state",
            "ef-normal",
            "categorical-state",
            "continuous-state",
            "binary-state",
            "generic-state",
        ]:
            raise ValueError(
                (
                    "Invalid node type. Should be one of the following: "
                    "'DP-state', 'continuous-state', 'binary-state', 'ef-normal'."
                    "'generic-state' or 'categorical-state'"
                )
            )

        # assess children number
        # this is required to ensure the coupling functions match
        children_number = 1
        if value_children is None:
            children_number = 0
        elif isinstance(value_children, int):
            children_number = 1
        elif isinstance(value_children, list):
            children_number = len(value_children)

        # transform coupling parameter into tuple of indexes and strenghts
        couplings = []
        for indexes in [
            value_parents,
            volatility_parents,
            value_children,
            volatility_children,
        ]:
            if indexes is not None:
                if isinstance(indexes, int):
                    coupling_idxs = tuple([indexes])
                    coupling_strengths = tuple([1.0])
                elif isinstance(indexes, list):
                    coupling_idxs = tuple(indexes)
                    coupling_strengths = tuple([1.0] * len(coupling_idxs))
                elif isinstance(indexes, tuple):
                    coupling_idxs = tuple(indexes[0])
                    coupling_strengths = tuple(indexes[1])
            else:
                coupling_idxs, coupling_strengths = None, None
            couplings.append((coupling_idxs, coupling_strengths))
        value_parents, volatility_parents, value_children, volatility_children = (
            couplings
        )

        # create the default parameters set according to the node type
        if kind == "continuous-state":
            default_parameters = {
                "mean": 0.0,
                "expected_mean": 0.0,
                "precision": 1.0,
                "expected_precision": 1.0,
                "volatility_coupling_children": volatility_children[1],
                "volatility_coupling_parents": volatility_parents[1],
                "value_coupling_children": value_children[1],
                "value_coupling_parents": value_parents[1],
                "tonic_volatility": -4.0,
                "tonic_drift": 0.0,
                "autoconnection_strength": 1.0,
                "observed": 1,
                "temp": {
                    "effective_precision": 0.0,
                    "value_prediction_error": 0.0,
                    "volatility_prediction_error": 0.0,
                },
            }
        elif kind == "binary-state":
            default_parameters = {
                "observed": 1,
                "mean": 0,
                "expected_mean": 0.5,
                "precision": 1.0,
                "expected_precision": 1.0,
                "value_coupling_parents": value_parents[1],
                "temp": {
                    "value_prediction_error": 0.0,
                },
            }
        elif kind == "generic-state":
            default_parameters = {
                "mean": 0.0,
                "observed": 1,
            }
        elif "ef-normal" in kind:
            default_parameters = {
                "nus": 3.0,
                "xis": jnp.array([0.0, 1.0]),
                "mean": 0.0,
                "observed": 1.0,
            }
        elif kind == "categorical-state":
            if "n_categories" in node_parameters:
                n_categories = node_parameters["n_categories"]
            elif "n_categories" in additional_parameters:
                n_categories = additional_parameters["n_categories"]
            else:
                n_categories = 4
            binary_parameters = {
                "n_categories": n_categories,
                "precision_1": 1.0,
                "precision_2": 1.0,
                "precision_3": 1.0,
                "mean_1": 1 / n_categories,
                "mean_2": -jnp.log(n_categories - 1),
                "mean_3": 0.0,
                "tonic_volatility_2": -4.0,
                "tonic_volatility_3": -4.0,
            }
            binary_idxs: List[int] = [
                1 + i + len(self.edges) for i in range(n_categories)
            ]
            default_parameters = {
                "binary_idxs": binary_idxs,  # type: ignore
                "n_categories": n_categories,
                "surprise": 0.0,
                "kl_divergence": 0.0,
                "alpha": jnp.ones(n_categories),
                "observed": jnp.ones(n_categories, dtype=int),
                "mean": jnp.array([1.0 / n_categories] * n_categories),
                "binary_parameters": binary_parameters,
            }
        elif kind == "DP-state":

            if "batch_size" in additional_parameters.keys():
                batch_size = additional_parameters["batch_size"]
            elif "batch_size" in node_parameters.keys():
                batch_size = node_parameters["batch_size"]
            else:
                batch_size = 10

            default_parameters = {
                "batch_size": batch_size,  # number of branches available in the network
                "n": jnp.zeros(batch_size),  # number of observation in each cluster
                "n_total": 0,  # the total number of observations in the node
                "alpha": 1.0,  # concentration parameter for the implied Dirichlet dist.
                "expected_means": jnp.zeros(batch_size),
                "expected_sigmas": jnp.ones(batch_size),
                "sensory_precision": 1.0,
                "activated": jnp.zeros(batch_size),
                "value_coupling_children": (1.0,),
                "mean": 0.0,
                "n_active_cluster": 0,
            }

        # Update the default node parameters using keywords args and dictonary
        if bool(additional_parameters):
            # ensure that all passed values are valid keys
            invalid_keys = [
                key
                for key in additional_parameters.keys()
                if key not in default_parameters.keys()
            ]

            if invalid_keys:
                raise ValueError(
                    (
                        "Some parameter(s) passed as keyword arguments were not found "
                        f"in the default key list for this node (i.e. {invalid_keys})."
                        " If you want to create a new key in the node attributes, "
                        "please use the node_parameters argument instead."
                    )
                )

            # if keyword parameters were provided, update the default_parameters dict
            default_parameters.update(additional_parameters)

        # update the defaults using the dict parameters
        default_parameters.update(node_parameters)
        node_parameters = default_parameters

        # define the type of node that is created
        if kind == "generic-state":
            node_type = 0
        elif kind == "binary-state":
            node_type = 1
        elif kind == "continuous-state":
            node_type = 2
        elif kind == "ef-normal":
            node_type = 3
        elif kind == "DP-state":
            node_type = 4
        elif kind == "categorical-state":
            node_type = 5

        for _ in range(n_nodes):
            # convert the structure to a list to modify it
            edges_as_list: List = list(self.edges)

            node_idx = len(self.edges)  # the index of the new node

            # for mutiple value children, set a default tuple with corresponding length
            if children_number != len(coupling_fn):
                if coupling_fn == (None,):
                    coupling_fn = children_number * coupling_fn
                else:
                    raise ValueError(
                        "The number of coupling fn and value children do not match"
                    )

            # add a new edge
            edges_as_list.append(
                AdjacencyLists(
                    node_type, None, None, None, None, coupling_fn=coupling_fn
                )
            )

            # convert the list back to a tuple
            self.edges = tuple(edges_as_list)

            # update the node structure
            self.attributes[node_idx] = deepcopy(node_parameters)

            # Update the edges of the parents and children accordingly
            # --------------------------------------------------------
            if value_parents[0] is not None:
                self.add_edges(
                    kind="value",
                    parent_idxs=value_parents[0],
                    children_idxs=node_idx,
                    coupling_strengths=value_parents[1],  # type: ignore
                )
            if value_children[0] is not None:
                self.add_edges(
                    kind="value",
                    parent_idxs=node_idx,
                    children_idxs=value_children[0],
                    coupling_strengths=value_children[1],  # type: ignore
                    coupling_fn=coupling_fn,
                )
            if volatility_children[0] is not None:
                self.add_edges(
                    kind="volatility",
                    parent_idxs=node_idx,
                    children_idxs=volatility_children[0],
                    coupling_strengths=volatility_children[1],  # type: ignore
                )
            if volatility_parents[0] is not None:
                self.add_edges(
                    kind="volatility",
                    parent_idxs=volatility_parents[0],
                    children_idxs=node_idx,
                    coupling_strengths=volatility_parents[1],  # type: ignore
                )

        if kind == "categorical-state":
            # if we are creating a categorical state or state-transition node
            # we have to generate the implied binary network(s) here
            self = fill_categorical_state_node(
                self,
                node_idx=node_idx,
                binary_states_idxs=node_parameters["binary_idxs"],  # type: ignore
                binary_parameters=binary_parameters,
            )

        return self

    def plot_nodes(self, node_idxs: Union[int, List[int]], **kwargs):
        """Plot the node(s) beliefs trajectories."""
        return plot_nodes(network=self, node_idxs=node_idxs, **kwargs)

    def plot_trajectories(self, **kwargs):
        """Plot the parameters trajectories."""
        return plot_trajectories(network=self, **kwargs)

    def plot_correlations(self):
        """Plot the heatmap of cross-trajectories correlation."""
        return plot_correlations(network=self)

    def plot_network(self):
        """Visualization of node network using GraphViz."""
        return plot_network(network=self)

    def to_pandas(self) -> pd.DataFrame:
        """Export the nodes trajectories and surprise as a Pandas data frame.

        Returns
        -------
        structure_df :
            Pandas data frame with the time series of sufficient statistics and
            the surprise of each node in the structure.

        """
        return to_pandas(self)

    def surprise(
        self,
        response_function: Callable,
        response_function_inputs: Tuple = (),
        response_function_parameters: Optional[
            Union[np.ndarray, ArrayLike, float]
        ] = None,
    ) -> float:
        """Surprise of the model conditioned by the response function.

        The surprise (negative log probability) depends on the input data, the model
        parameters, the response function, its inputs and its additional parameters
        (optional).

        Parameters
        ----------
        response_function :
            The response function to use to compute the model surprise. If `None`
            (default), return the sum of Gaussian surprise if `model_type=="continuous"`
            or the sum of the binary surprise if `model_type=="binary"`.
        response_function_inputs :
            A list of tuples with the same length as the number of models. Each tuple
            contains additional data and parameters that can be accessible to the
            response functions.
        response_function_parameters :
            A list of additional parameters that will be passed to the response
            function. This can include values over which inferece is performed in a
            PyMC model (e.g. the inverse temperature of a binary softmax).

        Returns
        -------
        surprise :
            The model's surprise given the input data and the response function.

        """
        return response_function(
            hgf=self,
            response_function_inputs=response_function_inputs,
            response_function_parameters=response_function_parameters,
        )
        return self

    def add_edges(
        self,
        kind="value",
        parent_idxs=Union[int, List[int]],
        children_idxs=Union[int, List[int]],
        coupling_strengths: Union[float, List[float], Tuple[float]] = 1.0,
        coupling_fn: Tuple[Optional[Callable], ...] = (None,),
    ) -> "Network":
        """Add a value or volatility coupling link between a set of nodes.

        Parameters
        ----------
        kind :
            The kind of coupling, can be `"value"` or `"volatility"`.
        parent_idxs :
            The index(es) of the parent node(s).
        children_idxs :
            The index(es) of the children node(s).
        coupling_strengths :
            The coupling strength betwen the parents and children.
        coupling_fn :
            Coupling function(s) between the current node and its value children.
            It has to be provided as a tuple. If multiple value children are specified,
            the coupling functions must be stated in the same order of the children.
            Note: if a node has multiple parents nodes with different coupling
            functions, a coupling function should be indicated for all the parent nodes.
            If no coupling function is stated, the relationship between nodes is assumed
            linear.

        """
        attributes, edges = add_edges(
            attributes=self.attributes,
            edges=self.edges,
            kind=kind,
            parent_idxs=parent_idxs,
            children_idxs=children_idxs,
            coupling_strengths=coupling_strengths,
            coupling_fn=coupling_fn,
        )

        self.attributes = attributes
        self.edges = edges

        return self



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

import jax.numpy as jnp

from pyhgf import load_data
from pyhgf.model import Network


def test_one_node_hgf():

    # one level HGF and one observation step
    one_node__hgf = (
        Network()
        .add_nodes()
        .add_nodes(value_children=0)
        .input_data(input_data=jnp.array([0.2]))
    )

    for key, val in zip(
        ["precision", "expected_precision", "mean", "expected_mean"],
        [1.0, 1.0, 0.2, 0.0],
    ):
        assert jnp.isclose(one_node__hgf.node_trajectories[0][key], val)
    for key, val in zip(
        ["precision", "expected_precision", "mean", "expected_mean"],
        [1.9820137, 0.98201376, 0.10090748, 0.0],
    ):
        assert jnp.isclose(one_node__hgf.node_trajectories[1][key], val)


def test_two_nodes_hgf():

    # one level HGF with a volatility parent and one observation step
    two_nodes__hgf = (
        Network()
        .add_nodes()
        .add_nodes(value_children=0)
        .add_nodes(volatility_children=0)
        .input_data(input_data=jnp.array([0.2]))
    )

    for key, val in zip(
        ["precision", "expected_precision", "mean", "expected_mean"],
        [1.0, 0.5, 0.2, 0.0],
    ):
        assert jnp.isclose(two_nodes__hgf.node_trajectories[0][key], val)
    for key, val in zip(
        ["precision", "expected_precision", "mean", "expected_mean"],
        [1.4820137, 0.98201376, 0.06747576, 0.0],
    ):
        assert jnp.isclose(two_nodes__hgf.node_trajectories[1][key], val)
    for key, val in zip(
        ["precision", "expected_precision", "mean", "expected_mean"],
        [1.1070137, 0.98201376, -0.12219789, 0.0],
    ):
        assert jnp.isclose(two_nodes__hgf.node_trajectories[2][key], val)


def test_nonlinear_coupling_fn():
    """Tests if the coupling function is passed correctly
    into the network"""

    # creating a simple coupling function
    def coupling_fn(x):
        return jnp.sin(x)

    def identity_fn(x):
        return x

    # I create a network with a node with 2 value children
    test_HGF = (
        Network()
        .add_nodes(n_nodes=2)
        .add_nodes(value_children=0, coupling_fn=(identity_fn,))
        .add_nodes(value_children=1, coupling_fn=(coupling_fn,))
    )

    # check if the number of coupling fn matches the number of children
    coupling_fn_length = []
    children_number = []
    for node_idx in range(2, len(test_HGF.edges)):
        coupling_fn_length.append(len(test_HGF.edges[node_idx].coupling_fn))
        children_number.append(len(test_HGF.edges[node_idx].value_children))

    assert children_number == coupling_fn_length

    # pass data
    test_HGF.input_data(input_data=jnp.array([0.2, 0.2]))

    for idx, val in zip(
        ["precision", "expected_precision", "mean", "expected_mean"],
        [2.9125834, 1.9125834, 0.13492969, 0.10090748],
    ):
        assert jnp.isclose(test_HGF.node_trajectories[2][idx][1], val)
    for idx, val in zip(
        ["precision", "expected_precision", "mean", "expected_mean"],
        [2.9125834, 1.9125834, 0.0, 0.0],
    ):
        assert jnp.isclose(test_HGF.node_trajectories[3][idx][1], val)


def test_continuous_scan_loop():
    timeserie = load_data("continuous")

    two_level_hgf = (
        Network()
        .add_nodes()
        .add_nodes(value_children=0)
        .add_nodes(volatility_children=0)
        .input_data(input_data=timeserie)
    )

    for idx, val in zip(
        ["precision", "expected_precision", "mean", "expected_mean"],
        [1.0, 0.9716732, 0.8241, 0.79489124],
    ):
        assert jnp.isclose(two_level_hgf.node_trajectories[0][idx][-1], val)

    for idx, val in zip(
        ["precision", "expected_precision", "mean", "expected_mean"],
        [7.785051, 6.813378, 0.79853684, 0.79489124],
    ):
        assert jnp.isclose(two_level_hgf.node_trajectories[1][idx][-1], val)

    for idx, val in zip(
        ["precision", "expected_precision", "mean", "expected_mean"],
        [0.25191233, 0.25114372, -3.5367591, -3.5352085],
    ):
        assert jnp.isclose(two_level_hgf.node_trajectories[2][idx][-1], val)



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from typing import Tuple, Union

import jax.numpy as jnp
from jax import Array
from jax.scipy.special import digamma, gamma, gammaln
from jax.typing import ArrayLike


class MultivariateNormal:
    """The multivariate normal as an exponential family distribution.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Exponential_family

    """

    @staticmethod
    def sufficient_statistics(x: ArrayLike) -> Array:
        """Compute the sufficient statistics for the multivariate normal."""
        return jnp.hstack([x, jnp.outer(x, x)[jnp.tril_indices(x.shape[0])]])

    @staticmethod
    def base_measure(k: int) -> float:
        """Compute the base measures for the multivariate normal."""
        return (2 * jnp.pi) ** (-k / 2)


class Normal:
    """The univariate normal as an exponential family distribution.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Exponential_family

    """

    @staticmethod
    def sufficient_statistics(x: float) -> Array:
        """Sufficient statistics for the univariate normal."""
        return jnp.array([x, x**2])

    @staticmethod
    def expected_sufficient_statistics(mu: float, sigma) -> Array:
        """Compute expected sufficient statistics from the mean and std."""
        return jnp.array([mu, mu**2 + sigma**2])

    @staticmethod
    def base_measure() -> float:
        """Compute the base measure of the univariate normal."""
        return 1 / (jnp.sqrt(2 * jnp.pi))

    @staticmethod
    def parameters(xis: ArrayLike) -> Tuple[float, float]:
        """Get parameters from the expected sufficient statistics.

        Parameters
        ----------
        xis :
            The expected sufficient statistics.

        Returns
        -------
        mean, variance :
            The parameters of the distribution (mean and variance).

        """
        mean = xis[0]
        variance = xis[1] - (mean**2)
        return mean, variance


def gaussian_predictive_distribution(x: float, xi: ArrayLike, nu: float) -> float:
    r"""Density of the Gaussian-predictive distribution.

    This distribution is parametrized by hyperparameters from the exponential family as:

    .. math::

        \begin{cases}
            \mathcal{NP}(x, \xi, \nu) :=
            \sqrt{\frac{1}{\pi(\nu+1)(\xi_{x^2}-\xi_{x}^2)}}
            \frac{\Gamma(\frac{\nu+2}{2})}{\Gamma(\frac{\nu+1}{2})}
            \left( 1+\frac{(x-\xi_{x})^2}{(\nu+1)(\xi_{x^2}-\xi_x^2)} \right)
            ^{-\frac{\nu+2}{2}}
        \end{cases}

    Parameters
    ----------
    x :
        The point at which the density is evaluated.
    xi :
        Hyperparameter updated by the sufficient statistics of the observed variables.
    nu :
        Hyperparameter over the number of valid observation (pseudo-counts).

    Returns
    -------
    y :
        The probability density evaluated at *x*.

    References
    ----------
    .. [1] Mathys, C., & Weber, L. (2020). Hierarchical Gaussian filtering of sufficient
       statistic time series for active inference. In Communications in Computer and
       Information Science. Active Inference (pp. 52–58).
       doi:10.1007/978-3-030-64919-7_7

    """
    return (
        jnp.sqrt(1 / (jnp.pi * (nu + 1) * (xi[1] - xi[0] ** 2)))
        * jnp.exp(
            (gammaln((nu + 2) / 2) - gammaln((nu + 1) / 2))
        )  # use gammaln to avoid numerical overflow
        * (1 + ((x - xi[0]) ** 2) / ((nu + 1) * (xi[1] - xi[0] ** 2)))
        ** (-(nu + 2) / 2)
    )


def gaussian_density(x: ArrayLike, mean: ArrayLike, precision: ArrayLike) -> ArrayLike:
    """Gaussian density as defined by mean and precision."""
    return precision / jnp.sqrt(2 * jnp.pi) * jnp.exp(-precision / 2 * (x - mean) ** 2)


def sigmoid(
    x: Union[ArrayLike, float],
    lower_bound: Union[ArrayLike, float] = 0.0,
    upper_bound: Union[ArrayLike, float] = 1.0,
) -> ArrayLike:
    """Logistic sigmoid function."""
    return (upper_bound - lower_bound) / (1 + jnp.exp(-x)) + lower_bound


def binary_surprise(
    x: Union[float, ArrayLike], expected_mean: Union[float, ArrayLike]
) -> ArrayLike:
    r"""Surprise at a binary outcome.

    The surprise ellicited by a binary observation :math:`x` under the expected
    probability :math:`\hat{\mu}` is given by:

    .. math::

       \begin{cases}
            -\log(\hat{\mu}),& \text{if } x=1\\
            -\log(1 - \hat{\mu}), & \text{if } x=0\\
        \end{cases}

    Parameters
    ----------
    x :
        The outcome.
    expected_mean :
        The mean of the Bernoulli distribution.

    Returns
    -------
    surprise :
        The binary surprise.


    Examples
    --------
    >>> from pyhgf.binary import binary_surprise
    >>> binary_surprise(x=1.0, expected_mean=0.7)
    `Array(0.35667497, dtype=float32, weak_type=True)`

    """
    return jnp.where(
        x, -jnp.log(expected_mean), -jnp.log(jnp.array(1.0) - expected_mean)
    )


def gaussian_surprise(
    x: Union[float, ArrayLike],
    expected_mean: Union[float, ArrayLike],
    expected_precision: Union[float, ArrayLike],
) -> Array:
    r"""Surprise at an outcome under a Gaussian prediction.

    The surprise elicited by an observation :math:`x` under a Gaussian distribution
    with expected mean :math:`\hat{\mu}` and expected precision :math:`\hat{\pi}` is
    given by:

    .. math::

       \frac{1}{2} (\log(2 \pi) - \log(\hat{\pi}) + \hat{\pi}(x - \hat{\mu})^2)

    where :math:`\pi` is the mathematical constant.

    Parameters
    ----------
    x :
        The outcome.
    expected_mean :
        The expected mean of the Gaussian distribution.
    expected_precision :
        The expected precision of the Gaussian distribution.

    Returns
    -------
    surprise :
        The Gaussian surprise.

    Examples
    --------
    >>> from pyhgf.math import gaussian_surprise
    >>> gaussian_surprise(x=2.0, expected_mean=0.0, expected_precision=1.0)
    `Array(2.9189386, dtype=float32, weak_type=True)`

    """
    return jnp.array(0.5) * (
        jnp.log(jnp.array(2.0) * jnp.pi)
        - jnp.log(expected_precision)
        + expected_precision * jnp.square(jnp.subtract(x, expected_mean))
    )


def dirichlet_kullback_leibler(alpha_1: ArrayLike, alpha_2: ArrayLike) -> Array:
    r"""Compute the Kullback-Leibler divergence between two Dirichlet distributions.

    The Kullback-Leibler divergence from the distribution :math:`Q` to the distribution
    :math:`P`, two Dirichlet distributions parametrized by :math:`\alpha_2` and
    :math:`\alpha_1` (respectively) is given by the following equation:

    .. math::
       KL[P||Q] = \ln{\frac{\Gamma(\sum_{i=1}^k\alpha_{1i})}
         {\Gamma(\sum_{i=1}^k\alpha_{2i})}} +
         \sum_{i=1}^k \ln{\frac{\Gamma(\alpha_{2i})}{\Gamma(\alpha_{1i})}} +
         \sum_{i=1}^k(\alpha_{1i} -
         \alpha_{2i})\left[\psi(\alpha_{1i})-\psi(\sum_{i=1}^k\alpha_{1i})\right]

    Parameters
    ----------
    alpha_1 :
        The concentration parameters for the distribution :math:`P`.
    alpha_2 :
        The concentration parameters for the distribution :math:`Q`.

    Returns
    -------
    kl :
        The Kullback-Leibler divergence of distribution :math:`P` from distribution
        :math:`Q`.

    References
    ----------
    .. [1] https://statproofbook.github.io/P/dir-kl.html
    .. [2] Penny, William D. (2001): "KL-Divergences of Normal, Gamma, Dirichlet and
       Wishart densities" ; in: University College, London , p. 2, eqs. 8-9 ;
       URL: https://www.fil.ion.ucl.ac.uk/~wpenny/publications/densities.ps .

    """
    return (
        jnp.log(gamma(alpha_1.sum()) / gamma(alpha_2.sum()))
        + jnp.sum(jnp.log(gamma(alpha_2) / gamma(alpha_1)))
        + jnp.sum((alpha_1 - alpha_2) * (digamma(alpha_1) - digamma(alpha_1.sum())))
    )


def binary_surprise_finite_precision(
    value: Union[ArrayLike, float],
    expected_mean: Union[ArrayLike, float],
    expected_precision: Union[ArrayLike, float],
    eta0: Union[ArrayLike, float] = 0.0,
    eta1: Union[ArrayLike, float] = 1.0,
) -> Array:
    r"""Compute the binary surprise with finite precision.

    Parameters
    ----------
    value :
        The observed value.
    expected_mean :
        The expected probability of observing category 1.
    expected_precision :
        The precision of the underlying normal distributions.
    eta0 :
        The first possible value.
    eta1 :
        The second possible value.

    Returns
    -------
    surprise :
        The binary surprise under finite precision.

    """
    return -jnp.log(
        expected_mean * gaussian_density(value, eta1, expected_precision)
        + (1 - expected_mean) * gaussian_density(value, eta0, expected_precision)
    )


def sigmoid_inverse_temperature(x: float, temperature: float) -> float:
    """Compute the sigmoid response function with inverse temperature parameter."""
    return (x**temperature) / (x**temperature + (1 - x) ** temperature)



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

import jax.numpy as jnp
import numpy as np

from pyhgf import load_data
from pyhgf.model import HGF, Network
from pyhgf.response import total_gaussian_surprise


def test_network():
    """Test the network class"""

    #####################
    # Creating networks #
    #####################

    custom_hgf = (
        Network()
        .add_nodes(kind="continuous-state")
        .add_nodes(kind="binary-state")
        .add_nodes(value_children=0)
        .add_nodes(
            value_children=1,
        )
        .add_nodes(value_children=[2, 3])
        .add_nodes(value_children=4)
        .add_nodes(volatility_children=[2, 3])
        .add_nodes(volatility_children=2)
        .add_nodes(volatility_children=7)
    )

    custom_hgf.create_belief_propagation_fn(overwrite=False)
    custom_hgf.create_belief_propagation_fn(overwrite=True)

    custom_hgf.input_data(input_data=np.ones((10, 2)))


def test_continuous_hgf():
    """Test the continuous HGF"""
    ##############
    # Continuous #
    ##############
    timeserie = load_data("continuous")

    # two-level
    # ---------
    two_level_continuous_hgf = HGF(
        n_levels=2,
        model_type="continuous",
        initial_mean={"1": timeserie[0], "2": 0.0},
        initial_precision={"1": 1e4, "2": 1e1},
        tonic_volatility={"1": -3.0, "2": -3.0},
        tonic_drift={"1": 0.0, "2": 0.0},
        volatility_coupling={"1": 1.0},
    )

    two_level_continuous_hgf.input_data(input_data=timeserie)

    surprise = two_level_continuous_hgf.surprise()  # Sum the surprise for this model
    assert jnp.isclose(surprise.sum(), -1141.0911)
    assert len(two_level_continuous_hgf.node_trajectories[1]["mean"]) == 614

    # three-level
    # -----------
    three_level_continuous_hgf = HGF(
        n_levels=3,
        model_type="continuous",
        initial_mean={"1": 1.04, "2": 1.0, "3": 1.0},
        initial_precision={"1": 1e4, "2": 1e1, "3": 1e1},
        tonic_volatility={"1": -13.0, "2": -2.0, "3": -2.0},
        tonic_drift={"1": 0.0, "2": 0.0, "3": 0.0},
        volatility_coupling={"1": 1.0, "2": 1.0},
    )
    three_level_continuous_hgf.input_data(input_data=timeserie)
    surprise = three_level_continuous_hgf.surprise()
    assert jnp.isclose(surprise.sum(), -892.82227)

    # test an alternative response function
    sp = total_gaussian_surprise(three_level_continuous_hgf)
    assert jnp.isclose(sp.sum(), -2545.4248)


def test_binary_hgf():
    """Test the binary HGF"""

    ##########
    # Binary #
    ##########
    u, _ = load_data("binary")

    # two-level
    # ---------
    two_level_binary_hgf = HGF(
        n_levels=2,
        model_type="binary",
        initial_mean={"1": 0.0, "2": 0.5},
        initial_precision={"1": 0.0, "2": 1e4},
        tonic_volatility={"1": None, "2": -6.0},
        tonic_drift={"1": None, "2": 0.0},
        volatility_coupling={"1": None},
        eta0=0.0,
        eta1=1.0,
        binary_precision=jnp.inf,
    )

    # Provide new observations
    two_level_binary_hgf = two_level_binary_hgf.input_data(u)
    surprise = two_level_binary_hgf.surprise()
    assert jnp.isclose(surprise.sum(), 215.58821)

    # three-level
    # -----------
    three_level_binary_hgf = HGF(
        n_levels=3,
        model_type="binary",
        initial_mean={"1": 0.0, "2": 0.5, "3": 0.0},
        initial_precision={"1": 0.0, "2": 1e4, "3": 1e1},
        tonic_volatility={"1": None, "2": -6.0, "3": -2.0},
        tonic_drift={"1": None, "2": 0.0, "3": 0.0},
        volatility_coupling={"1": None, "2": 1.0},
        eta0=0.0,
        eta1=1.0,
        binary_precision=jnp.inf,
    )
    three_level_binary_hgf.input_data(input_data=u)
    surprise = three_level_binary_hgf.surprise()
    assert jnp.isclose(surprise.sum(), 215.59067)


def test_custom_sequence():
    """Test the continuous HGF"""

    ############################
    # dynamic update sequences #
    ############################
    u, _ = load_data("binary")

    three_level_binary_hgf = HGF(
        n_levels=3,
        model_type="binary",
        initial_mean={"1": 0.0, "2": 0.5, "3": 0.0},
        initial_precision={"1": 0.0, "2": 1e4, "3": 1e1},
        tonic_volatility={"1": None, "2": -6.0, "3": -2.0},
        tonic_drift={"1": None, "2": 0.0, "3": 0.0},
        volatility_coupling={"1": None, "2": 1.0},
        eta0=0.0,
        eta1=1.0,
        binary_precision=jnp.inf,
    )

    # create a custom update series
    update_sequence1 = three_level_binary_hgf.update_sequence
    update_sequence2 = update_sequence1[:2]
    update_branches = (update_sequence1, update_sequence2)
    branches_idx = np.random.binomial(n=1, p=0.5, size=len(u))

    three_level_binary_hgf.input_custom_sequence(
        update_branches=update_branches,
        branches_idx=branches_idx,
        input_data=u,
    )



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

import jax.numpy as jnp
import numpy as np

from pyhgf.model import Network


def test_categorical_state_node():
    # generate some categorical inputs data
    np.random.seed(123)
    input_data = np.array(
        [np.random.multinomial(n=1, pvals=[0.1, 0.2, 0.7]) for _ in range(3)]
    ).T
    input_data = np.vstack([[0.0] * input_data.shape[1], input_data])

    # create the categorical HGF
    categorical_hgf = Network().add_nodes(
        kind="categorical-state",
        node_parameters={
            "n_categories": 3,
            "binary_parameters": {"tonic_volatility_2": -2.0},
        },
    )

    # fitting the model forwards
    categorical_hgf.input_data(
        input_data=(input_data, np.ones(input_data.shape, dtype=int))
    )

    # export to pandas data frame
    categorical_hgf.to_pandas()

    assert jnp.isclose(
        categorical_hgf.node_trajectories[0]["kl_divergence"].sum(), 0.8222526
    )
    assert jnp.isclose(
        categorical_hgf.node_trajectories[0]["surprise"].sum(), 7.6738853
    )



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

import arviz as az
import jax.numpy as jnp
import numpy as np
import pymc as pm
from jax import grad, jit, vmap
from jax.tree_util import Partial

from pyhgf import load_data
from pyhgf.distribution import (
    HGFDistribution,
    HGFLogpGradOp,
    HGFPointwise,
    hgf_logp,
    logp,
)
from pyhgf.model import HGF
from pyhgf.response import (
    binary_softmax_inverse_temperature,
    first_level_gaussian_surprise,
)


def test_logp():
    """Test the log-probability function for single model fit."""
    timeseries = load_data("continuous")
    hgf = HGF(n_levels=2, model_type="continuous")

    log_likelihood = logp(
        tonic_volatility_1=-3.0,
        tonic_volatility_2=-3.0,
        tonic_volatility_3=jnp.nan,
        input_precision=np.array(1e4),
        tonic_drift_1=0.0,
        tonic_drift_2=0.0,
        tonic_drift_3=jnp.nan,
        precision_1=1e4,
        precision_2=1e1,
        precision_3=jnp.nan,
        mean_1=1.0,
        mean_2=0.0,
        mean_3=jnp.nan,
        volatility_coupling_1=1.0,
        volatility_coupling_2=jnp.nan,
        response_function_parameters=[jnp.nan],
        input_data=timeseries,
        time_steps=np.ones(shape=timeseries.shape),
        response_function_inputs=jnp.nan,
        response_function=first_level_gaussian_surprise,
        hgf=hgf,
    )

    assert jnp.isclose(log_likelihood.sum(), 1141.0585)


def test_vectorized_continuous_logp():
    """Test the vectorized version of the log-probability function."""

    timeseries = load_data("continuous")
    hgf = HGF(n_levels=2, model_type="continuous")

    # generate data with 2 parameters set
    input_data = np.array([timeseries] * 2)
    time_steps = np.ones(shape=input_data.shape)

    tonic_volatility_1 = -3.0
    tonic_volatility_2 = -3.0
    tonic_volatility_3 = jnp.nan
    input_precision = np.array(1e4)
    tonic_drift_1 = 0.0
    tonic_drift_2 = 0.0
    tonic_drift_3 = jnp.nan
    precision_1 = 1e4
    precision_2 = 1e1
    precision_3 = jnp.nan
    mean_1 = 1.0
    mean_2 = 0.0
    mean_3 = jnp.nan
    volatility_coupling_1 = 1.0
    volatility_coupling_2 = jnp.nan
    response_function_parameters = jnp.ones(2)

    # Broadcast inputs to an array with length n>=1
    (
        _tonic_volatility_1,
        _tonic_volatility_2,
        _tonic_volatility_3,
        _input_precision,
        _tonic_drift_1,
        _tonic_drift_2,
        _tonic_drift_3,
        _precision_1,
        _precision_2,
        _precision_3,
        _mean_1,
        _mean_2,
        _mean_3,
        _volatility_coupling_1,
        _volatility_coupling_2,
        _,
    ) = jnp.broadcast_arrays(
        tonic_volatility_1,
        tonic_volatility_2,
        tonic_volatility_3,
        input_precision,
        tonic_drift_1,
        tonic_drift_2,
        tonic_drift_3,
        precision_1,
        precision_2,
        precision_3,
        mean_1,
        mean_2,
        mean_3,
        volatility_coupling_1,
        volatility_coupling_2,
        jnp.zeros(2),
    )

    # create the vectorized version of the function
    vectorized_logp_two_levels = vmap(
        Partial(
            logp,
            hgf=hgf,
            response_function=first_level_gaussian_surprise,
        )
    )

    # model fit
    log_likelihoods = vectorized_logp_two_levels(
        input_data=input_data,
        response_function_inputs=jnp.ones(2),
        response_function_parameters=response_function_parameters,
        time_steps=time_steps,
        mean_1=_mean_1,
        mean_2=_mean_2,
        mean_3=_mean_3,
        precision_1=_precision_1,
        precision_2=_precision_2,
        precision_3=_precision_3,
        tonic_volatility_1=_tonic_volatility_1,
        tonic_volatility_2=_tonic_volatility_2,
        tonic_volatility_3=_tonic_volatility_3,
        tonic_drift_1=_tonic_drift_1,
        tonic_drift_2=_tonic_drift_2,
        tonic_drift_3=_tonic_drift_3,
        volatility_coupling_1=_volatility_coupling_1,
        volatility_coupling_2=_volatility_coupling_2,
        input_precision=_input_precision,
    )

    assert jnp.isclose(log_likelihoods.sum(), 2282.1165).all()


def test_vectorized_binary_logp():
    """Test the vectorized version of the log-probability function."""

    u, y = load_data("binary")
    hgf = HGF(n_levels=2, model_type="binary")

    # generate data with 2 parameters set
    input_data = np.array([u] * 2)
    time_steps = np.ones(shape=input_data.shape)

    tonic_volatility_1 = -3.0
    tonic_volatility_2 = -3.0
    tonic_volatility_3 = jnp.nan
    input_precision = jnp.nan
    tonic_drift_1 = 0.0
    tonic_drift_2 = 0.0
    tonic_drift_3 = jnp.nan
    precision_1 = jnp.nan
    precision_2 = 1.0
    precision_3 = jnp.nan
    mean_1 = 0.5
    mean_2 = 0.0
    mean_3 = jnp.nan
    volatility_coupling_1 = 1.0
    volatility_coupling_2 = jnp.nan
    response_function_parameters = jnp.ones(2)

    # Broadcast inputs to an array with length n>=1
    (
        _tonic_volatility_1,
        _tonic_volatility_2,
        _tonic_volatility_3,
        _input_precision,
        _tonic_drift_1,
        _tonic_drift_2,
        _tonic_drift_3,
        _precision_1,
        _precision_2,
        _precision_3,
        _mean_1,
        _mean_2,
        _mean_3,
        _volatility_coupling_1,
        _volatility_coupling_2,
        _,
    ) = jnp.broadcast_arrays(
        tonic_volatility_1,
        tonic_volatility_2,
        tonic_volatility_3,
        input_precision,
        tonic_drift_1,
        tonic_drift_2,
        tonic_drift_3,
        precision_1,
        precision_2,
        precision_3,
        mean_1,
        mean_2,
        mean_3,
        volatility_coupling_1,
        volatility_coupling_2,
        jnp.zeros(2),
    )

    # create the vectorized version of the function
    vectorized_logp_two_levels = vmap(
        Partial(
            logp,
            hgf=hgf,
            response_function=binary_softmax_inverse_temperature,
        )
    )

    # model fit
    log_likelihoods = vectorized_logp_two_levels(
        input_data=input_data,
        response_function_inputs=np.array([y] * 2),
        response_function_parameters=response_function_parameters,
        time_steps=time_steps,
        mean_1=_mean_1,
        mean_2=_mean_2,
        mean_3=_mean_3,
        precision_1=_precision_1,
        precision_2=_precision_2,
        precision_3=_precision_3,
        tonic_volatility_1=_tonic_volatility_1,
        tonic_volatility_2=_tonic_volatility_2,
        tonic_volatility_3=_tonic_volatility_3,
        tonic_drift_1=_tonic_drift_1,
        tonic_drift_2=_tonic_drift_2,
        tonic_drift_3=_tonic_drift_3,
        volatility_coupling_1=_volatility_coupling_1,
        volatility_coupling_2=_volatility_coupling_2,
        input_precision=_input_precision,
    )

    assert jnp.isclose(log_likelihoods.sum(), -283.83594).all()


def test_hgf_logp():
    """Test the hgf_logp function used by Distribution Ops on three level models"""

    ##############################
    # Three-level continuous HGF #
    ##############################
    timeseries = load_data("continuous")
    continuous_hgf = HGF(n_levels=3, model_type="continuous")

    # generate data with 2 parameters set
    input_data = np.array([timeseries] * 2)
    time_steps = np.ones(shape=input_data.shape)

    # create the vectorized version of the function
    vectorized_logp_three_levels = vmap(
        Partial(
            logp,
            hgf=continuous_hgf,
            response_function=first_level_gaussian_surprise,
        )
    )

    sum_log_likelihoods, log_likelihoods = hgf_logp(
        tonic_volatility_1=-3.0,
        tonic_volatility_2=-3.0,
        tonic_volatility_3=-3.0,
        input_precision=1e4,
        tonic_drift_1=0.0,
        tonic_drift_2=0.0,
        tonic_drift_3=0.0,
        precision_1=1e4,
        precision_2=1e1,
        precision_3=1.0,
        mean_1=1.0,
        mean_2=0.0,
        mean_3=0.0,
        volatility_coupling_1=1.0,
        volatility_coupling_2=1.0,
        response_function_parameters=np.ones(2),
        response_function_inputs=np.ones(2),
        vectorized_logp=vectorized_logp_three_levels,
        input_data=input_data,
        time_steps=time_steps,
    )
    assert sum_log_likelihoods == log_likelihoods.sum()
    assert jnp.isclose(sum_log_likelihoods, 2269.6929).all()

    # test the gradient
    grad_logp = jit(
        grad(
            Partial(
                hgf_logp,
                vectorized_logp=vectorized_logp_three_levels,
                input_data=input_data,
                time_steps=time_steps,
                response_function_inputs=np.ones(2),
            ),
            argnums=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15],
            has_aux=True,
        ),
    )

    gradients, _ = grad_logp(
        1.0,
        0.0,
        0.0,
        1.0,
        1.0,
        1.0,
        -3.0,
        -3.0,
        -3.0,
        0.0,
        0.0,
        0.0,
        1.0,
        1.0,
        1e4,
        np.ones(2),
    )
    assert jnp.isclose(gradients[0], 0.09576362)
    assert jnp.isclose(gradients[1], -8.531818)
    assert jnp.isclose(gradients[2], -189.85936)

    ##########################
    # Three-level binary HGF #
    ##########################

    # Create the data (value and time vectors)
    u, y = load_data("binary")
    binary_hgf = HGF(n_levels=3, model_type="binary")

    # generate data with 2 parameters set
    input_data = np.array([u] * 2)
    response_function_inputs = np.array([y] * 2)
    time_steps = np.ones(shape=input_data.shape)

    # create the vectorized version of the function
    vectorized_logp_three_levels = vmap(
        Partial(
            logp,
            hgf=binary_hgf,
            response_function=binary_softmax_inverse_temperature,
        )
    )

    sum_log_likelihoods, log_likelihoods = hgf_logp(
        vectorized_logp=vectorized_logp_three_levels,
        tonic_volatility_1=np.nan,
        tonic_volatility_2=-2.0,
        tonic_volatility_3=-6.0,
        precision_1=np.nan,
        precision_2=1.0,
        precision_3=1.0,
        mean_1=0.5,
        mean_2=0.0,
        mean_3=0.0,
        volatility_coupling_1=1.0,
        volatility_coupling_2=1.0,
        response_function_parameters=np.array([1.0, 1.0]),
        input_data=input_data,
        response_function_inputs=response_function_inputs,
        time_steps=time_steps,
    )

    assert sum_log_likelihoods == log_likelihoods.sum()
    assert jnp.isclose(sum_log_likelihoods, -248.07889)

    # test the gradient
    grad_logp = jit(
        grad(
            Partial(
                hgf_logp,
                vectorized_logp=vectorized_logp_three_levels,
                input_data=input_data,
                time_steps=time_steps,
                response_function_inputs=response_function_inputs,
            ),
            argnums=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15],
            has_aux=True,
        ),
    )

    gradients, _ = grad_logp(
        0.5,
        0.0,
        0.0,
        1.0,
        1.0,
        1.0,
        -3.0,
        -2.0,
        -6.0,
        0.0,
        0.0,
        0.0,
        1.0,
        1.0,
        np.inf,
        np.ones(2),
    )
    assert jnp.isclose(gradients[0], 0.0)
    assert jnp.isclose(gradients[1], 1.9314771)
    assert jnp.isclose(gradients[2], 30.185408)


def test_pytensor_pointwise_logp():
    """Test the pytensor HGFPointwise op."""

    ##############
    # Binary HGF #
    ##############

    # Create the data (value and time vectors)
    u, y = load_data("binary")

    hgf_logp_op = HGFPointwise(
        input_data=u[np.newaxis, :],
        model_type="binary",
        n_levels=2,
        response_function=binary_softmax_inverse_temperature,
        response_function_inputs=y[np.newaxis, :],
    )

    logp = hgf_logp_op(
        tonic_volatility_1=np.inf,
        tonic_volatility_2=-6.0,
        tonic_volatility_3=np.inf,
        input_precision=np.inf,
        tonic_drift_1=0.0,
        tonic_drift_2=0.0,
        tonic_drift_3=np.inf,
        precision_1=0.0,
        precision_2=1e4,
        precision_3=np.inf,
        mean_1=np.inf,
        mean_2=0.5,
        mean_3=np.inf,
        volatility_coupling_1=1.0,
        volatility_coupling_2=np.inf,
        response_function_parameters=np.array([1.0]),
    ).eval()

    assert jnp.isclose(logp.sum(), -200.2442167699337)


def test_pytensor_logp():
    """Test the pytensor hgf_logp op."""

    ##################
    # Continuous HGF #
    ##################

    # Create the data (value and time vectors)
    timeseries = load_data("continuous")

    hgf_logp_op = HGFDistribution(
        input_data=timeseries[np.newaxis, :],
        model_type="continuous",
        n_levels=2,
        response_function=first_level_gaussian_surprise,
        response_function_inputs=None,
    )

    logp = hgf_logp_op(
        mean_1=np.array(1.0),
        mean_2=np.array(0.0),
        mean_3=np.array(0.0),
        precision_1=np.array(1e4),
        precision_2=np.array(1e1),
        precision_3=np.array(0.0),
        tonic_volatility_1=np.array(-3.0),
        tonic_volatility_2=np.array(-3.0),
        tonic_volatility_3=np.array(0.0),
        tonic_drift_1=np.array(0.0),
        tonic_drift_2=np.array(0.0),
        tonic_drift_3=np.array(0.0),
        volatility_coupling_1=np.array(1.0),
        volatility_coupling_2=np.array(0.0),
        input_precision=np.array(1e4),
        response_function_parameters=np.ones(1),
    ).eval()

    assert jnp.isclose(logp, 1141.05847168)

    ##############
    # Binary HGF #
    ##############

    # Create the data (value and time vectors)
    u, y = load_data("binary")

    hgf_logp_op = HGFDistribution(
        input_data=u[np.newaxis, :],
        model_type="binary",
        n_levels=2,
        response_function=binary_softmax_inverse_temperature,
        response_function_inputs=y[np.newaxis, :],
    )

    logp = hgf_logp_op(
        tonic_volatility_1=np.inf,
        tonic_volatility_2=-6.0,
        tonic_volatility_3=np.inf,
        input_precision=np.inf,
        tonic_drift_1=0.0,
        tonic_drift_2=0.0,
        tonic_drift_3=np.inf,
        precision_1=0.0,
        precision_2=1e4,
        precision_3=np.inf,
        mean_1=np.inf,
        mean_2=0.5,
        mean_3=np.inf,
        volatility_coupling_1=1.0,
        volatility_coupling_2=np.inf,
        response_function_parameters=np.array([1.0]),
    ).eval()

    assert jnp.isclose(logp, -200.2442167699337)


def test_pytensor_grad_logp():
    """Test the pytensor gradient hgf_logp op."""

    ##################
    # Continuous HGF #
    ##################

    # Create the data (value and time vectors)
    timeseries = load_data("continuous")

    hgf_logp_grad_op = HGFLogpGradOp(
        model_type="continuous",
        input_data=timeseries[np.newaxis, :],
        n_levels=2,
        response_function=first_level_gaussian_surprise,
        response_function_inputs=None,
    )

    tonic_volatility_1 = hgf_logp_grad_op(
        tonic_volatility_1=-3.0,
        tonic_volatility_2=-3.0,
        input_precision=1e4,
        tonic_drift_1=0.0,
        tonic_drift_2=0.0,
        precision_1=1.0,
        precision_2=1.0,
        mean_1=1.0,
        mean_2=0.0,
        volatility_coupling_1=1.0,
    )[6].eval()

    assert jnp.isclose(tonic_volatility_1, -3.3479857)

    ##############
    # Binary HGF #
    ##############

    # Create the data (value and time vectors)
    u, y = load_data("binary")

    hgf_logp_grad_op = HGFLogpGradOp(
        model_type="binary",
        input_data=u[np.newaxis, :],
        n_levels=2,
        response_function=binary_softmax_inverse_temperature,
        response_function_inputs=y[np.newaxis, :],
    )

    tonic_volatility_2 = hgf_logp_grad_op(
        tonic_volatility_1=jnp.inf,
        tonic_volatility_2=-6.0,
        input_precision=jnp.inf,
        tonic_drift_1=0.0,
        tonic_drift_2=0.0,
        precision_1=0.0,
        precision_2=1e4,
        mean_1=jnp.inf,
        mean_2=0.5,
        volatility_coupling_1=1.0,
    )[7].eval()

    assert jnp.isclose(tonic_volatility_2, 10.866466)


def test_pymc_sampling():
    """Test the pytensor hgf_logp op."""

    ##############
    # Continuous #
    ##############

    # Create the data (value and time vectors)
    timeseries = load_data("continuous")

    hgf_logp_op = HGFDistribution(
        n_levels=2,
        input_data=timeseries[np.newaxis, :],
        response_function=first_level_gaussian_surprise,
    )

    with pm.Model() as model:
        tonic_volatility_2 = pm.Uniform("tonic_volatility_2", -10, 0)

        pm.Potential(
            "hhgf_loglike",
            hgf_logp_op(
                tonic_volatility_2=tonic_volatility_2,
                input_precision=np.array(1e4),
            ),
        )

    initial_point = model.initial_point()

    pointslogs = model.point_logps(initial_point)
    assert pointslogs["tonic_volatility_2"] == -1.39
    assert pointslogs["hhgf_loglike"] == 1468.28

    with model:
        idata = pm.sample(chains=2, cores=1, tune=1000)

    assert -9.5 > az.summary(idata)["mean"].values[0] > -10.5
    assert az.summary(idata)["r_hat"].values[0] <= 1.02

    ##########
    # Binary #
    ##########

    # Create the data (value and time vectors)
    u, y = load_data("binary")

    hgf_logp_op = HGFDistribution(
        n_levels=2,
        model_type="binary",
        input_data=u[np.newaxis, :].repeat(2, axis=0),
        response_function=binary_softmax_inverse_temperature,
        response_function_inputs=y[np.newaxis, :].repeat(2, axis=0),
    )

    def logp(value, tonic_volatility_2):
        return hgf_logp_op(tonic_volatility_2=tonic_volatility_2)

    with pm.Model() as model:
        y_data = pm.Data("y_data", y[np.newaxis, :].repeat(2, axis=0))
        tonic_volatility_2 = pm.Normal("tonic_volatility_2", -11.0, 2, shape=(2,))
        pm.CustomDist("likelihood", tonic_volatility_2, logp=logp, observed=y_data)

    initial_point = model.initial_point()

    pointslogs = model.point_logps(initial_point)
    assert pointslogs["tonic_volatility_2"] == -3.22
    assert pointslogs["likelihood"] == -425.18

    with model:
        idata = pm.sample(chains=2, cores=1, tune=1000)

    assert -2 < round(az.summary(idata)["mean"].values[0]) < 0



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from functools import partial
from typing import Callable, Dict

import jax.numpy as jnp
from jax import jit

from pyhgf.typing import Attributes, Edges


@partial(jit, static_argnames=("edges", "node_idx", "sufficient_stats_fn"))
def posterior_update_exponential_family(
    attributes: Dict, edges: Edges, node_idx: int, sufficient_stats_fn: Callable, **args
) -> Attributes:
    r"""Update the parameters of an exponential family distribution.

    Assuming that :math:`nu` is fixed, updating the hyperparameters of the distribution
    is given by:

    .. math::
        \xi \leftarrow \xi + \frac{1}{\nu + 1}(t(x)-\xi)

    Parameters
    ----------
    attributes :
        The attributes of the probabilistic nodes.
    edges :
        The edges of the probabilistic nodes as a tuple of
        :py:class:`pyhgf.typing.Indexes`. The tuple has the same length as the node
        number. For each node, the index lists the value and volatility parents and
        children.
    node_idx :
        Pointer to the value parent node that will be updated.
    sufficient_stats_fn :
        Compute the sufficient statistics of the probability distribution. This should
        be one of the method implemented in the distribution class in
        :py:class:`pyhgf.math.Normal`, for a univariate normal.

    Returns
    -------
    attributes :
        The updated attributes of the probabilistic nodes.

    References
    ----------
    .. [1] Mathys, C., & Weber, L. (2020). Hierarchical Gaussian Filtering of Sufficient
       Statistic Time Series for Active Inference. In Active Inference (pp. 52–58).
       Springer International Publishing. https://doi.org/10.1007/978-3-030-64919-7_7

    """
    # update the hyperparameter vectors
    xis = attributes[node_idx]["xis"] + (1 / (1 + attributes[node_idx]["nus"])) * (
        sufficient_stats_fn(x=attributes[node_idx]["mean"])
        - attributes[node_idx]["xis"]
    )

    # blank update in the case of unobserved value
    attributes[node_idx]["xis"] = jnp.where(
        attributes[node_idx]["observed"], xis, attributes[node_idx]["xis"]
    )

    return attributes
